"""
Spectral energy distributions
===============================

Methods and classes for dealing with spectral energy distributions (SEDs).
"""

import numpy
from . import lsd # intrapackage reference

class SED:
	"""
	Class that defines objects storing Spectral energy distributions (SEDs). 
	The SEDs can be imported either from datafiles with the observations/models,
	or from arrays with the values of log(nu) and log(nuLnu).
	
	The object's initial attributes are:

	- nu: frequency in Hertz
	- lognu: log10(frequency)
	- nlnu: nu L_nu in erg/s
	- ll: log10(nu Lnu)
	- file : input datafile
	
	To import a SED from an ASCII file:

	>>> s=sed.SED(file=datafile, logfmt=XX)

	where logfmt is 1 if the data is already in logarithm or 0 otherwise.
	NOTE: logfmt for now is by default 0 (Eracleous' SEDs).
	
	To create a SED object from the arrays lognu and lognulnu:

	>>> s=sed.SED(lognu=lognu, ll=lognulnu)
	
	To import a SED in a special format:

	>>> s=sed.SED()
	>>> s.erac('ngc1097.dat')
	
	To plot the SED:

	>>> plot(s.lognu, s.ll)
	
	To calculate the bolometric luminosity:

	>>> s.bol()
	
	after which the value will be stored in s.lumbol	
	"""
		
	def __init__(self, file=None, logfmt=0, lognu=None, ll=None):
		"""
	To import a SED from an ASCII file:
	>>> s=sed.SED(file=XX, logfmt=XX)
	where logfmt is 1 if the data is already in logarithm or 0 otherwise.
	NOTE: logfmt for now is by default 0 (e.g. Eracleous' SEDs).
	
	To create a SED object from the arrays lognu and lognulnu:

	>>> s=sed.SED(lognu=lognu, ll=lognulnu)
	
	>>> s=sed.SED()

	creates an empty SED object. Then you can call specific methods to handle
	SEDs in special formats (e.g. s.erac).
		"""
		self.file=file	# input datafile (string) from which the observed spectrum was imported
		self.lognu=lognu
		self.ll=ll

		# If the SED is imported from a file...
		if file!=None:
			if logfmt==0:
				# Reads SED from datafile
				self.nu,self.nlnu = numpy.loadtxt(file,unpack=True,usecols=(0,1))
		
				# Take logs
				self.lognu, self.ll = numpy.log10(self.nu), numpy.log10(self.nlnu)
			else:
				self.lognu, self.ll = numpy.loadtxt(file,unpack=True,usecols=(0,1))
				self.nu,self.nlnu = 10.**self.lognu, 10.**self.ll

		# If the SED is created from the arrays
		if file==None and lognu!=None:
			if logfmt==1:
				self.lognu, self.ll = lognu, ll
				self.nu,self.nlnu = 10.**self.lognu, 10.**self.ll
			else:
				self.nu, self.nlnu = lognu, ll
				self.lognu,self.ll = numpy.log10(self.lognu), numpy.log10(self.ll)
		
		# Checks if ll has NaN or Inf values
		if file!=None or lognu!=None:
			self.check()



	def erac(self, file):
		"""
		Reads SEDs in the format provided by Eracleous et al. LINER SEDs:
		nu	nuLnu	nuLnu(extinction)	error	upperLimit?(0/1)
		
		Adds the following attributes to the SED object:
		- nlnuex: nuLnu after extinction correction
		- llex: log10(nuLnu) after extinction correction
		- ul: is the datapoint an upper limit (arrow)? 1 if yes, 0 if not
		"""
		self.file=file
		
		self.nu,self.nlnu,self.nlnuex,self.ul = numpy.loadtxt(file,unpack=True,usecols=(0,1,2,4))
		self.lognu, self.ll = numpy.log10(self.nu), numpy.log10(self.nlnu)
		self.llex = numpy.log10(self.nlnuex)
		
		# Checks if ll has NaN or Inf values
		self.check()




	def hayden(self, file, dist):
		"""
		Reads SEDs in the format provided by Hampadarath et al.:
	nu(GHz) flux(microJy) error(microJy) Detection[1]orUpperlimit[0]
		
		Adds the following attributes to the SED object:
		- ul: is the datapoint an upper limit (arrow)? 1 if yes, 0 if not
		- distance: distance to object in Mpc
		"""
		self.file=file
		self.distance=dist
		
		self.nu,flux,x,self.ul = numpy.loadtxt(file,unpack=True,usecols=(0,1,2,3))
		flux=flux*1e-26*1e-6*1e7*1e-4	# microJy to erg/s/cm^2/Hz
		dist=dist*3.086e24	# Mpc to cm
		self.lnu=4.*numpy.pi*dist**2*flux
		self.lognu=numpy.log10(self.nu)
		self.ll=numpy.log10(self.lnu)+self.lognu
		

		# inverts the convention for upper limits, to be consistent with my
		# previous one (Hayden adopts the inverse convention)
		x=self.ul.copy()
		x[self.ul==0]=1
		x[self.ul==1]=0
		self.ul=x

		# Checks if ll has NaN or Inf values
		self.check()



	def prieto(self, file, dist):
		"""
		Reads SEDs in the format provided by Hampadarath et al.:
	nu(Hz) nu*F_nu(Jy Hz) 
		
		Adds the following attributes to the SED object:
		- distance: distance to object in Mpc
		"""
		self.file=file
		self.distance=dist
		
		self.lognu,nfn = numpy.loadtxt(file,unpack=True,usecols=(0,1),delimiter=',')
		self.nu=10**self.lognu
		nfn=10**nfn*1e-26*1e7*1e-4	# Jy to erg/s/cm^2
		dist=dist*3.086e24	# Mpc to cm
		self.ll=numpy.log10(4.*numpy.pi*dist**2*nfn)
		self.lnu=10**self.ll/self.nu

		# Checks if ll has NaN or Inf values
		self.check()



	def grmonty(self, file):
		"""
		Reads SEDs in the format provided by grmonty. I ported the 
		original SM script `plspec.m` provided by J. Dolence. 
		"""
		import astropy.io.ascii as ascii

		# reads grmonty SED file
		self.file=file
		s=ascii.read(file, Reader=ascii.NoHeader)

		# carries out array conversions
		small = 1e-12
		ll0 = numpy.log10(s['col2']+small) + numpy.log10(3.83e33)
		ll1 = numpy.log10(s['col8']+small) + numpy.log10(3.83e33)
		ll2 = numpy.log10(s['col14']+small) + numpy.log10(3.83e33)
		ll3 = numpy.log10(s['col20']+small) + numpy.log10(3.83e33)
		ll4 = numpy.log10(s['col26']+small) + numpy.log10(3.83e33)
		ll5 = numpy.log10(s['col32']+small) + numpy.log10(3.83e33)
		lw = s['col1'] + numpy.log10(9.1e-28*3e10*3e10/6.626e-27)

		# gets log(nu) and log(nuLnu)
		self.lognu, self.ll = lw, ll5
		self.nu,self.nlnu = 10**self.lognu,10**self.ll
		
		# Checks if ll has NaN or Inf values
		self.check()








	def haydenx(self, file):
		"""
		Reads X-ray data in the format provided by F. Panessa:
		nu	error_nu nuLnu	error	?
		
		Adds the following attributes to the SED object:
		- ul: is the datapoint an upper limit (arrow)? 1 if yes, 0 if not
		"""
		self.file=file
		
		self.lognu,self.lognuerr,self.ll,self.llerr = numpy.loadtxt(file,unpack=True,usecols=(0,1,2,4))
		self.lognuerr=numpy.abs(self.lognu-self.lognuerr)
		self.llerr=numpy.abs(self.ll-self.llerr)
		self.ul=numpy.zeros_like(self.ll)
		
		# Checks if ll has NaN or Inf values
		self.check()










	def check(self):
		"""
	Checks whether there are NaN or Inf values in the array log10(nuLnu).
	If so, displays a warning to the user and the corresponding filename.
	Method called by initialization methods init and erac.
	
	Creates a new attribute, weird, which is 1 if there are NaN or Inf values,
	and 0 otherwise.
		"""
		if False in numpy.isfinite(self.ll):
			print('Warning: NaN or Inf in the Lnu array! '),
			print(self.file)
			self.weird=1
		else:
			self.weird=0








	def sort(self):
		"""
	>>> s.sort()

	Sorts the SED in ascending order of frequencies.
		"""
		# Index of sorted elements
		i=self.lognu.argsort()
		self.nu,self.nlnu = self.nu[i],self.nlnu[i]
		self.lognu,self.ll = self.lognu[i],self.ll[i]
		
		# Precaution in case the user used the interp method
		if hasattr(self, 'nlnui'): 
			self.nui,self.nlnui = self.nui[i],self.nlnui[i]
			self.lognui,self.lli = self.lognui[i],self.lli[i]
		

		
	def unit(self):
		"""Normalizes the spectrum to one."""

		# Precaution in case the user did not use the interp method
		if not hasattr(self, 'nlnui'): self.interp()
	
		self.nlnu=self.nlnu/self.nlnu.max()
		self.ll=numpy.log10(self.nlnu)
		
		# Normalizes the interpolated values
		self.nlnui=self.nlnui/self.nlnui.max()
		self.lli=numpy.log10(self.nlnui)






	def write(self, file):
		"""
		Exports the SED as an ASCII file with the structure:
		col1=log10(nu/Hz), col2=log10(nu Lnu / erg per sec)
		
		>>> s.write(file)

		exports the SED s as the file "file"
		"""
		numpy.savetxt(file,numpy.transpose((self.lognu,self.ll)))






	def interp(self, seds=None, points=1000, xrange=[8,22]):
		"""
	Interpolates the SED and a list of SEDs (optional) in the given range
	of values of log(nu) and for the given number of points.
	
	>>> s.interp([s,s1,s2])
	interpolates s, s1 and s2 with 1000 points in the range
	log(nu)=8-22 (Hz). s, s1 and s2 are also instances of the class SED. The
	interpolated arrays (lognui, lli, nui, nlnui) will be added as new 
	attributes to each object. If given no argument, will interpolate the
	current SED object.
	
	:param seds: list of SED objects to interpolate
	:param points: number of interpolated points
	:param xrange: list in the form [xinitial, xfinal] with the x-range of interpolation. 
	
	If provided with no list of SEDs (e.g., s.interp()), then the method
	interpolates only the current SED s. If provided with no xrange argument,
	it will by default assume the range [8,22] discarding data points outside
	that range in the interpolation, hence adjust accordingly.
		"""
		# If seds==None, then the method operates only on the object itself 
		# (e.g., s.interp()) and proper actions are taken to avoid problems
		if seds==None:
			# Defines the new array of interpolated frequencies (with a 
			# preferentially large number of points)
			xold, yold = self.lognu, self.ll
			xnew=numpy.linspace(xrange[0],xrange[-1],points)
			ynew=numpy.interp(xnew,xold,yold,left=-20,right=-20)

			# Interpolates the current SED (simple linear interpolation).
			# Interpolated y-values outside the original x-range are zeroes (-20
			# in log space of luminosities).			
			self.lognui, self.lli = xnew, ynew	# log scale
			self.nui, self.nlnui = 10.**xnew, 10.**ynew
		else:
			# Interpolates the other SEDs with the same binning. But only does
			# that if a list of SEDs was actually provided!
			for sed in seds:
				xold, yold = sed.lognu, sed.ll
				xnew=numpy.linspace(xrange[0],xrange[-1],points)
				ynew=numpy.interp(xnew,xold,yold,left=-20,right=-20)
			
				sed.lognui, sed.lli = xnew, ynew	# log scale
				sed.nui, sed.nlnui = 10.**xnew, 10.**ynew




		



	
	
	
	def normalize(self, seds=None, nuref=17.684, refnlnu=1e40, xray=None):
		"""
	Normalizes a SED at the given frequency (lognu/Hz) and nuLnu (erg/s). 
	NOTE: this will modify the SED(s). Alternatively, if the user provides 
	the argument xray=True, then the method normalizes the SEDs such that 
	they have the same X-ray luminosity refnlnu in the range 2-10 keV 
	(thus ignoring the arg.	nuref).
	
	>>> s.normalize([s1,s2,s3])

	normalizes the SEDs in the object list at the same frequency and nuLnu
	
	>>> s.normalize([s1,s2,s3],refnlnu=1e40,xray=True)

	normalizes the SEDs such that they have the same X-ray luminosity 1e40 
	erg/s in the range 2-10 keV.
	
	The method is designed to automatically call the interp method (e.g., 
	s.interp([s1,s2])) if needed.
		"""
		# Precaution in case the user did not use the interp method
		if not hasattr(self, 'nlnui'): self.interp(seds)
		
		# If seds==None, then the method operates on the object itself, 
		# otherwise it operates on a list of objects
		if seds==None:
			if xray==None:	# if xray keyword is not provided
				# Finds the nuLnu corresponding to the frequency nearest nuref.
				# Uses the interpolated arrays for this purpose.
				i=lsd.search(nuref, self.lognui)
			
				self.nlnu=self.nlnu*refnlnu/self.nlnui[i]
				self.nlnui=self.nlnui*refnlnu/self.nlnui[i]
				self.lli=numpy.log10(self.nlnui)
				self.ll=numpy.log10(self.nlnu)
			else:	# if xray keyword IS provided
				self.xrays()
				factor=refnlnu/self.lumx	# normalization factor
				
				self.nlnu=self.nlnu*factor
				self.nlnui=self.nlnui*factor
				self.lli=numpy.log10(self.nlnui)
				self.ll=numpy.log10(self.nlnu)
		else:
			for sed in seds:
				if xray==None:	# if xray keyword is not provided
					i=lsd.search(nuref, sed.lognui)
			
					sed.nlnu=sed.nlnu*refnlnu/sed.nlnui[i]
					sed.nlnui=sed.nlnui*refnlnu/sed.nlnui[i]
					sed.lli=numpy.log10(sed.nlnui)
					sed.ll=numpy.log10(sed.nlnu)
				else:	# if xray keyword IS provided
					sed.xrays()
					factor=refnlnu/sed.lumx	# normalization factor
				
					sed.nlnu=sed.nlnu*factor
					sed.nlnui=sed.nlnui*factor
					sed.lli=numpy.log10(sed.nlnui)
					sed.ll=numpy.log10(sed.nlnu)
		

		
		
		
		
	def meanlin(self, seds, nuref=17.684, refnlnu=1e40):
		"""
	Given a list of SEDs and a reference value of log10(nu), normalizes the 
	SEDs (s.normalize) at the given nuref and refnlnu, calculates the average 
	of their luminosities (log10(average nuLnu)) and returns the list [mean,sd] 
	where 'mean' is SED object with the mean and 'sd' is the object with the
	standard deviation.
	
	>>> mean=s.meanlin([s1,s2],17.684,1e40)
	returns mean <- [lognu, <s1,s2>],
	where <s,s1,s2> -> log10[<nuLnu(s),nuLnu(s1),nuLnu(s2)>], lognu being the common 
	units of frequency for the SEDs after interpolation.
		"""	
		self.normalize(seds,nuref,refnlnu)	# normalize them to a common nu
		
		s=self.sum(seds)
		lognu=s.lognu
		m=s.nlnu/len(seds)	# takes the average
		
		# Computes the standard deviation
		sd=numpy.zeros_like(m)	# initializes the sd array		
		for sed in seds:
			sd=sd+(sed.nlnui-m)**2
		sd=numpy.sqrt(sd/len(seds))
		
		meansed=SED(lognu=lognu, ll=numpy.log10(m), logfmt=1)
		sdsed=SED(lognu=lognu, ll=numpy.log10(sd), logfmt=1)
		
		return [meansed,sdsed]		
		#return SED(lognu=lognu, ll=numpy.log10(m))
		





		
	def mean(self, seds, nuref=17.684, refnlnu=1e40, xray=None):
		"""
	Given a list of SEDs and a reference value of log10(nu), normalizes the 
	SEDs (s.normalize) at the given nuref and refnlnu, calculates the average 
	of their luminosities (average(log10(nuLnu)), like Eracleous et al. 2010 does) 
	and returns the list [mean,sd] where 'mean' is SED object with the mean and 
	'sd' is the object with the	standard deviation.
	
	Alternatively, if the user provides the argument xray=True, then
	the method normalizes the SEDs such that they have the same X-ray
	luminosity refnlnu in the range 2-10 keV (thus ignoring the arg.
	nuref).
	
	>>> means=s.mean([s1,s2],17.684,1e40)

	returns mean <- [lognu, <s1,s2>],
	where <s,s1,s2> -> <log10[nuLnu(s),nuLnu(s1),nuLnu(s2)]>, lognu being the common 
	units of frequency for the SEDs after interpolation.
	
	>>> means=s.mean([s1,s2],refnlnu=1e40,xray=True)

	returns the mean after normalizing the SEDs to the X-ray lum.
	1e40 erg/s in the range 2-10 keV.
		"""
		self.normalize(seds,nuref,refnlnu,xray)	# normalize them to a common nu
			
		# Sum the values of log10(nuLnu)
		sums=numpy.zeros_like(seds[0].lli)	# initializes the sum
		for sed in seds:
			sums=sums+sed.lli

		lognu=seds[0].lognui		
		m=sums/len(seds)	# takes the average
		
		# Computes the standard deviation
		sd=numpy.zeros_like(sums)	# initializes the std. dev. array
		for sed in seds:
			sd=sd+(sed.lli-m)**2
		sd=numpy.sqrt(sd/len(seds))
		
		meansed=SED(lognu=lognu, ll=m, logfmt=1)
		sdsed=SED(lognu=lognu, ll=sd, logfmt=1)
		
		return [meansed,sdsed]
		#return SED(lognu=lognu, ll=m)





	def median(self, seds, nuref=17.684, refnlnu=1e40, xray=None):
		"""
	Given a list of SEDs and a reference value of log10(nu), normalizes the 
	SEDs (s.normalize) at the given nuref and refnlnu, calculates the average 
	of their luminosities (average(log10(nuLnu)), like Eracleous et al. 2010 does) 
	and returns the list [mean,sd] where 'mean' is SED object with the median and 
	'sd' is the object with the	standard deviation.
	
	Alternatively, if the user provides the argument xray=True, then
	the method normalizes the SEDs such that they have the same X-ray
	luminosity refnlnu in the range 2-10 keV (thus ignoring the arg.
	nuref).
	
	>>> means=s.median([s1,s2],17.684,1e40)

	returns median <- [lognu, <s1,s2>],
	where <s,s1,s2> -> <log10[nuLnu(s),nuLnu(s1),nuLnu(s2)]>, lognu being the common 
	units of frequency for the SEDs after interpolation.
	
	>>> means=s.median([s1,s2],refnlnu=1e40,xray=True)

	returns the mean after normalizing the SEDs to the X-ray lum.
	1e40 erg/s in the range 2-10 keV.
		"""
		self.normalize(seds,nuref,refnlnu,xray)	# normalize them to a common nu
		
		# Array that stores the median of luminosities
		m=numpy.zeros_like(seds[0].lli)
		
		# Loop that uses the numpy.median method
		for i in numpy.arange(m.size):
			x=numpy.zeros_like(numpy.array(seds))
			for j in numpy.arange(numpy.size(seds)):
				x[j]=seds[j].lli[i]
			m[i]=numpy.median(x)

		lognu=seds[0].lognui

		# Computes the standard deviation
		sd=numpy.zeros_like(m)	# initializes the std. dev. array
		for sed in seds:
			sd=sd+(sed.lli-m)**2
		sd=numpy.sqrt(sd/len(seds))
		
		meansed=SED(lognu=lognu, ll=m, logfmt=1)
		sdsed=SED(lognu=lognu, ll=sd, logfmt=1)
		
		return [meansed,sdsed]






	




	def geomean(self, seds, nuref=17.684, refnlnu=1e40, xray=None):
		"""
	Given a list of SEDs and a reference value of log10(nu), normalizes the 
	SEDs (s.normalize) at the given nuref and refnlnu, calculates the geometric
	mean of their luminosities (average(log10(nuLnu)), like Eracleous et al. 2010 does) 
	and returns the list [mean,sd] where 'mean' is SED object with the geometric 
	mean and 'sd' is the object with the standard deviation.
	
	Alternatively, if the user provides the argument xray=True, then
	the method normalizes the SEDs such that they have the same X-ray
	luminosity refnlnu in the range 2-10 keV (thus ignoring the arg.
	nuref).
	
	>>> means=s.mean([s1,s2],17.684,1e40)

	returns mean <- [lognu, <s1,s2>],
	where <s,s1,s2> -> <log10[nuLnu(s),nuLnu(s1),nuLnu(s2)]>, lognu being the common 
	units of frequency for the SEDs after interpolation.
	
	>>> means=s.mean([s1,s2],refnlnu=1e40,xray=True)
	
	returns the mean after normalizing the SEDs to the X-ray lum.
	1e40 erg/s in the range 2-10 keV.
		"""	
		self.normalize(seds,nuref,refnlnu,xray)	# normalize them to a common nu
		
		# Calculates the product of log(nuLnu) (produtorio)
		sums=seds[0].lli.copy()	# initializes the sum
		for i in numpy.arange(1,len(seds)):
			sums=sums*seds[i].lli

		lognu=seds[0].lognui		
		m=sums**(1./len(seds)) 	# takes the nth root
		
		# Computes the standard deviation
		sd=sums.copy()	# initializes the std. dev. array
		for sed in seds:
			sd=sd+(sed.lli-m)**2
		sd=numpy.sqrt(sd/len(seds))
		
		meansed=SED(lognu=lognu, ll=m, logfmt=1)
		sdsed=SED(lognu=lognu, ll=sd, logfmt=1)
		
		return [meansed,sdsed]







	
	
	
	def bol(self, lognu0=None, lognu1=None, xrange=[8,22]):
		"""
	Calculates the bolometric luminosity of the SED. Performs the integration 
	using the trapezoidal rule. Adds the attribute lumbol to the object with
	the result.
	
	>>> s.bol()

	makes the variable s.lumbol hold the bolometric luminosity.
	
	>>> s.bol(17, 20)

	makes s.lumbol hold the luminosity integrated in the range 1e17-1e20 Hz.

	>>> s.bol(18, 20, [s.lognu[0],s.lognu[-1]])

	gives the integrated luminosity in the range 1e18-1e20 Hz. Before integrating,
	interpolates the SED over its original frequency range with 1000 points.

		"""
		import scipy.integrate

		# Performs interpolation before integrating. This is a precaution in 
		# case the user specifies weird integration limits.	In addition, I 
		# found out that integrating certain sparse SEDs (e.g. qsos) without
		# interpolating induces incorrect Lbol estimates		
		if not hasattr(self, 'nlnui'): self.interp(xrange=xrange)			
		
		# If no arguments are given to the method, integrates the entire SED
		if lognu0==None and lognu1==None:
			lumbol=scipy.integrate.trapz(self.nlnui/self.nui, self.nui)
		else:
			# Gets only the elements in the given range
			i=numpy.where((self.lognui>=lognu0) & (self.lognui<=lognu1))
			
			# Integration
			lumbol=scipy.integrate.trapz(self.nlnui[i]/self.nui[i], self.nui[i])

		self.lumbol=lumbol

		return lumbol





	def edd(self,mass=None):
		"""
	Computes the Eddington ratio :math:`L_{\\rm bol}/L_{\\rm} Edd}`.
	First computes the bolometric luminosity via interpolation and 
	then calculates the Eddington ratio.

	:param mass: log10(BH mass in solar masses)
		"""
		# Computes Lbol before anything else	
		if not hasattr(self, 'lumbol'): 			
			lumbol=self.bol()

		# if no mass argument was given
		if mass==None: 
			mass=self.mass

		# Eddington ratio
		self.lumedd=1.3e38*10**mass
		self.eddratio=lumbol/self.lumedd

		return self.eddratio

	


	def xrays(self):
		"""
	Calculates the photon index of the X-ray spectrum and Lx (2-10 keV). 
	Stores the new attributes gammax and lumx in the object.
	
	>>> Lx,gamma=s.xrays()

	creates the new attributes s.lumx and s.gammax
		"""
		import scipy.integrate
		import scipy.stats
		
		# Performs interpolation before integrating. This is a precaution in 
		# case the user specify weird integration limits.			
		if not hasattr(self, 'nlnui'): self.interp()
		
		# 2-10 keV in log(nu)
		xi, xf = 17.684, 18.384
		
		# Gets only the elements in the range 2-10 keV
		i=numpy.where((self.lognui>=xi) & (self.lognui<=xf))
		x,y = self.lognui[i],self.lli[i]	# easier notation
		
		# Calculates Lx using integration (trapezoidal rule)
		self.lumx=scipy.integrate.trapz(self.nlnui[i]/self.nui[i], self.nui[i])

		# Linear fit to the SED in the interval 2-10 keV, y=ax+b
		a, b, r, p, err = scipy.stats.linregress(x,y)
		
		# Calculates Gamma_X (photon index)
		self.gammax=2-a

		return self.lumx, self.gammax
		
		
		
		
		
	
	def findlum(self,x):
		"""
	Given the input frequency in units of log(nu/Hz), returns the list [log(nu),
	Lnu] with the frequency and luminosity closest to the specified value.
	
	>>> lognu,lum=s.findlum(14)

	will look for the freq. and lum. nearest nu=10^14 Hz
		"""
		# Performs interpolation before the search
		if not hasattr(self, 'nlnui'): self.interp()
		
		# Looks for the frequency
		i=lsd.search(x,self.lognui)	# index
		
		return [self.lognui[i], self.nlnui[i]/self.nui[i]]
	
	
	
	
	
	
	def radioloud(self):
		"""
	Computes the radio-loudness for the SED. Adds the attribute rl to the 
	object with	the result. If Ro>=10. we have a RL AGN, otherwise it is a
	RQ one.
	
	>>> rl=s.radioloud()

	makes the variable s.rl hold the radio-loudness.
		"""
		# Radio luminosity at 6 cm. 6 cm -> 4.997x10^9 Hz -> lognu=9.7
		lumrad=self.findlum(9.7)[1]
		
		# Optical lum. at 4400 A. B band -> 4400 A -> 6.813x10^14 Hz -> lognu=14.833 
		lumopt=self.findlum(14.833)[1]
				
		# Kellermann recipe: Lnu(6 cm)/Lnu(B)
		return lumrad/lumopt





	def alphaox(self):
		"""
	Computes the alpha_ox index.

	>>> alphaox=s.alphaox()
		"""
		# Lnu(2 keV)
		lumx=self.findlum(17.684)[1]
		
		# UV lum. at 2500 AA -> 1.199e15 Hz  -> lognu=15.079 
		lumuv=self.findlum(15.079)[1]
		
		return numpy.log10(lumx/lumuv)/2.605





	def ion(self):
		"""
	Calculates the rate of ionizing photons in the SED.
	
	>>> q=s.ion()
		"""
		import scipy.integrate
		import scipy.stats
		h=6.62607e-27	# Planck constant in CGS
		
		# Performs interpolation before integrating. This is a precaution in 
		# case the user specify weird integration limits.			
		if not hasattr(self, 'nlnui'): self.interp()
		
		# 13.6 eV - "infty"
		xi, xf = 15.52, 22.
		
		# Gets only the elements corresponding to ionizing frequencies
		i=numpy.where((self.lognui>=xi) & (self.lognui<=xf))
		x,y = self.lognui[i],self.lli[i]	# easier notation
		
		# Calculates ionizing rate using integration (trapezoidal rule)
		q=scipy.integrate.trapz(self.nlnui[i]/self.nui[i]/(h*self.nui[i]), self.nui[i])

		return q


		
		
	
	
	
	def chisq(self, model, npars=None):
		"""
	Computes the goodness of fit between the observed SED (assumed to be "self") 
	and the given model. If you don't provide the number of free 
	parameters in the model, it returns a modified chi squared (modified in 
	the sense that it assumes the uncertainties in all data points to be unity).
	If you provide npars it returns the modified reduced chi squared.
	
	>>> s.chisq(m)
	returns the (modified) chi square with s and m being the observed and model
	SEDs, respectively.
	
	>>> s.chisq(m,4)
	returns the reduced chi square.
	
	NOTE: this method was written assuming that the observed SEDs correspond
	to Eracleous' SEDs. Needs to be adapted in order to be applied to other 
	types of SEDs. We discard the 100 keV interpolation done by EHF10.
		"""
		obs=self
		
		# Discards upper limits in the data and the 100 keV interpolated point 
		i=numpy.where((obs.ul==0) & (obs.lognu<19.))	# index of elements corresponding to good elements
		
		# Interpolates the model SED in the same binning as the obs.
		yobs=obs.ll[i]
		ymod=numpy.interp(obs.lognu[i],model.lognu,model.ll)

		# Chi-square 
		if npars==None:
			chisq=numpy.sum((yobs-ymod)**2)
		else:	# or reduced chi-square statistic
			nu=yobs.size-1-npars		# Number of degrees of freedom
			chisq=numpy.sum((yobs-ymod)**2)/nu
			
		return chisq
	
	
	
	
	
	def copy(self):
		"""
	Returns a copy of the current SED object. Use this in the same way you would use
	the numpy copy() method to avoid changing the attributes of a SED
	unintentionally.
		"""
		c=SED(lognu=self.lognu,ll=self.ll,file=self.file)
		
		# Precaution if SED was imported using erac method
		if hasattr(self, 'llex'):
			c.llex=self.llex
			c.nlnuex=self.nlnuex
			c.ul=self.ul
		
		return c




def sum(seds):
	"""
Given a list of SEDs previously interpolated in the same binning, 
sums their luminosities and returns a SED object with the sum.
	
>>> ss=sum([s,s1,s2])

returns the SED ss <- [lognu, s+s1+s2],
where s+s1+s2 -> log10[nuLnu(s)+nuLnu(s1)+nuLnu(s2)], lognu being the common 
units of frequency for the SEDs after interpolation.
	
The method is designed to automatically call the interp method for each SED
if needed.
	"""
	# Precaution in case the user did not use the interp method
	seds[0].interp(seds)
		
	sums=numpy.zeros_like(seds[0].lognui)	# initializes the sum

	for sed in seds:
		sums=sums+sed.nlnui
		
	return SED(lognu=seds[0].lognui, ll=numpy.log10(sums), logfmt=1)

		







def haydensed(source, patho='/Users/nemmen/work/projects/hayden/all/', info='/Users/nemmen/work/projects/hayden/info.dat'):
	"""
Reads an observed SED from the Hayden et al. sample. Computes useful 
quantities.

:param source: source name e.g. 'NGC1097' or 'ngc1097' or 'NGC 1097'. Must have four numbers.
:param patho: path to observed SED data files
:param info: path to information datafile with distances and masses

:returns: SED object 
	"""
	import astropy.io.ascii as ascii
	# READS SEDS
	# =============
	# Finds the required parameters to plot the data for the specified source.
	# First converts the source name to uppercase, no spaces
	source=source.upper()
	source=source.replace(' ', '')

	# Reads information table and gets required info
	t=ascii.read(info)
	names=t['name'].tolist()
	# finds location of source in the arrays
	i=names.index(source)
	# gets distance 
	dist=t['distance/Mpc'][i]
	# gets BH mass
	mass=t['logmass'][i]

	# Reads observed SED from radio to X-rays 
	# (which is given in weird units)
	ofile=patho+source+'_sed.txt'
	o=SED()
	o.hayden(ofile,dist)
	o.mass=mass	
		
	# X-ray data
	# Reads Panessa's X-ray SED and creates another SED object
	xfile=patho+source+'_x.txt'
	x=SED()
	x.haydenx(xfile)	

	# Stacks together all the multiwavelength information
	o.lognu=numpy.concatenate([o.lognu,x.lognu])
	o.ll=numpy.concatenate([o.ll,x.ll])	

	return o




"""
Methods for dealing with Bayesian statistics
==============================================

e.g. priors, posteriors, joint density plots.

Right now the module is focused around PyMC, but I am migrating
to emcee.

.. todo:: plot confidence/credibility interval of a model
"""

import numpy, pylab, scipy, scipy.stats





def joint_density(X, Y, bounds=None):
	"""
Plots joint distribution of variables.
Inherited from method in src/graphics.py module in project 
git://github.com/aflaxman/pymc-example-tfr-hdi.git
	"""
	if bounds:
		X_min, X_max, Y_min, Y_max = bounds
	else:
		X_min = X.min()
		X_max = X.max()
		Y_min = Y.min()
		Y_max = Y.max()

	pylab.plot(X, Y, linestyle='none', marker='o', color='green', mec='green', alpha=.2, zorder=-99)

	gkde = scipy.stats.gaussian_kde([X, Y])
	x,y = pylab.mgrid[X_min:X_max:(X_max-X_min)/25.,Y_min:Y_max:(Y_max-Y_min)/25.]
	z = pylab.array(gkde.evaluate([x.flatten(), y.flatten()])).reshape(x.shape)
	pylab.contour(x, y, z, linewidths=2)

	pylab.axis([X_min, X_max, Y_min, Y_max])







def allplot(xb,yb,bins=30,fig=1,xlabel='x',ylabel='y'):
	"""
Input:
X,Y : objects referring to the variables produced by PyMC that you want
to analyze. Example: X=M.theta, Y=M.slope.

Inherited from Tommy LE BLANC's code at astroplotlib|STSCI.
	"""
	#X,Y=xb.trace(),yb.trace()
	X,Y=xb,yb

	#pylab.rcParams.update({'font.size': fontsize})
	fig=pylab.figure(fig)
	pylab.clf()
	
	gs = pylab.GridSpec(2, 2, width_ratios=[3,1], height_ratios=[1,3], wspace=0.07, hspace=0.07)
	scat=pylab.subplot(gs[2])
	histx=pylab.subplot(gs[0], sharex=scat)
	histy=pylab.subplot(gs[3], sharey=scat)
	#scat=fig.add_subplot(2,2,3)
	#histx=fig.add_subplot(2,2,1, sharex=scat)
	#histy=fig.add_subplot(2,2,4, sharey=scat)
	
	# Scatter plot
	scat.plot(X, Y,linestyle='none', marker='o', color='green', mec='green',alpha=.2, zorder=-99)

	gkde = scipy.stats.gaussian_kde([X, Y])
	x,y = numpy.mgrid[X.min():X.max():(X.max()-X.min())/25.,Y.min():Y.max():(Y.max()-Y.min())/25.]
	z = numpy.array(gkde.evaluate([x.flatten(), y.flatten()])).reshape(x.shape)
	scat.contour(x, y, z, linewidths=2)
	scat.set_xlabel(xlabel)
	scat.set_ylabel(ylabel)

	# X-axis histogram
	histx.hist(X, bins, histtype='stepfilled')
	pylab.setp(histx.get_xticklabels(), visible=False)	# no X label
	#histx.xaxis.set_major_formatter(pylab.NullFormatter())	# no X label

	# Y-axis histogram
	histy.hist(Y, bins, histtype='stepfilled', orientation='horizontal')
	pylab.setp(histy.get_yticklabels(), visible=False)	# no Y label
	#histy.yaxis.set_major_formatter(pylab.NullFormatter())	# no Y label

	#pylab.minorticks_on()
	#pylab.subplots_adjust(hspace=0.1)
	#pylab.subplots_adjust(wspace=0.1)
	pylab.draw()
	pylab.show()









def jointplotx(X,Y,xlabel=None,ylabel=None,binsim=40,binsh=20,binscon=15):
	"""
Plots the joint distribution of posteriors for X1 and X2, including the 1D
histograms showing the median and standard deviations. Uses simple method
for drawing the confidence contours compared to jointplot (which is wrong).

The work that went in creating this method is shown, step by step, in 
the ipython notebook "error contours.ipynb". Sources of inspiration:
- http://python4mpia.github.io/intro/quick-tour.html

Usage:
>>> jointplot(M.rtr.trace(),M.mdot.trace(),xlabel='$\log \ r_{\\rm tr}$', ylabel='$\log \ \dot{m}$')
	"""
	# Generates 2D histogram for image
	histt, xt, yt = numpy.histogram2d(X, Y, bins=[binsim,binsim], normed=False)
	histt = numpy.transpose(histt)  # Beware: numpy switches axes, so switch back.

	# assigns correct proportions to subplots
	fig=pylab.figure()
	gs = pylab.GridSpec(2, 2, width_ratios=[3,1], height_ratios=[1,3], wspace=0.001, hspace=0.001)
	con=pylab.subplot(gs[2])
	histx=pylab.subplot(gs[0], sharex=con)
	histy=pylab.subplot(gs[3], sharey=con)
		
	# Image
	con.imshow(histt,extent=[xt[0],xt[-1], yt[0],yt[-1]],origin='lower',cmap=pylab.cm.gray_r,aspect='auto')

	# Overplot with error contours 1,2 sigma
	# Contour plot
	histdata, x, y = numpy.histogram2d(X, Y, bins=[binscon,binscon], normed=False)
	histdata = numpy.transpose(histdata)  # Beware: numpy switches axes, so switch back.
	pmax  = histdata.max()
	cs=con.contour(histdata, levels=[0.68*pmax,0.05*pmax], extent=[x[0],x[-1], y[0],y[-1]], colors=['black','blue'])
	# use dictionary in order to assign your own labels to the contours.
	#fmtdict = {s[0]:r'$1\sigma$',s[1]:r'$2\sigma$'}
	#con.clabel(cs, fmt=fmtdict, inline=True, fontsize=20)
	if xlabel!=None: con.set_xlabel(xlabel)
	if ylabel!=None: con.set_ylabel(ylabel)

	# X-axis histogram
	histx.hist(X, binsh, histtype='stepfilled',facecolor='lightblue')
	pylab.setp(histx.get_xticklabels(), visible=False)	# no X label
	pylab.setp(histx.get_yticklabels(), visible=False)	# no Y label
	# Vertical lines with median and 1sigma confidence
	yax=histx.set_ylim()
	histx.plot([numpy.median(X),numpy.median(X)],[yax[0],yax[1]],'k-',linewidth=2) # median
	xsd=scipy.stats.scoreatpercentile(X, [15.87,84.13])
	histx.plot([xsd[0],xsd[0]],[yax[0],yax[1]],'k--') # -1sd
	histx.plot([xsd[-1],xsd[-1]],[yax[0],yax[1]],'k--') # +1sd

	# Y-axis histogram
	histy.hist(Y, binsh, histtype='stepfilled', orientation='horizontal',facecolor='lightyellow')
	pylab.setp(histy.get_yticklabels(), visible=False)	# no Y label
	pylab.setp(histy.get_xticklabels(), visible=False)	# no X label
	# Vertical lines with median and 1sigma confidence
	xax=histy.set_xlim()
	histy.plot([xax[0],xax[1]],[numpy.median(Y),numpy.median(Y)],'k-',linewidth=2) # median
	ysd=scipy.stats.scoreatpercentile(Y, [15.87,84.13])
	histy.plot([xax[0],xax[1]],[ysd[0],ysd[0]],'k--') # -1sd
	histy.plot([xax[0],xax[1]],[ysd[-1],ysd[-1]],'k--') # +1sd




"""
Statistical methods
=====================

- fit residuals
- Computing prediction and confidence bands
- Comparing goodness-of-fit of different models
- operations on statistical distributions
- custom statistical distributions
- p-values and significance
"""

import numpy,scipy
import scipy.stats




# Residuals of a fit
# ===================
# Residual sum of squares, raw scatter about best-fit and 
# intrinsic scatter.

def scatterfit(x,y,a=None,b=None):
	"""
Compute the mean deviation of the data about the linear model given if A,B
(*y=ax+b*) provided as arguments. Otherwise, compute the mean deviation about 
the best-fit line.

:param x,y: assumed to be Numpy arrays. 
:param a,b: scalars.
:rtype: float sd with the mean deviation.
	"""

	if a==None:	
		# Performs linear regression
		a, b, r, p, err = scipy.stats.linregress(x,y)
	
	# Std. deviation of an individual measurement (Bevington, eq. 6.15)
	N=numpy.size(x)
	sd=1./(N-2.)* numpy.sum((y-a*x-b)**2)
	sd=numpy.sqrt(sd)
	
	return sd
	

def scatterfitg(ydata,ymod,deg):
	"""
Compute the mean deviation of the data about the model given in
the array ymod, tabulated exactly like ydata.

Usage:

>>> sd=scatterfitg(ydata,ymod,n)

:param ydata: data
:param ymod: model evaluated at xdata
:param n: number of free parameters in the model
  
:type ydata,ymod: Numpy arrays
:rtype: float sd with the mean deviation.
	"""	
	# Std. deviation of an individual measurement (Bevington, eq. 6.15)
	N=numpy.size(ydata)
	sd=1./(N-deg)* numpy.sum((ydata-ymod)**2)
	sd=numpy.sqrt(sd)
	
	return sd



def scatpratt(x,y,errx,erry,a,b):
	"""
This is an alternative way of computing the "raw" scatter about the best-fit 
linear relation in the presence of measurement errors, proposed by 
Pratt et al. 2009, A&A, 498, 361. 

In the words of Cavagnolo et al. (2010): 
*We have quantified the total scatter about the best-fit relation using a 
weighted estimate of the orthogonal distances to the best-fit line (see 
Pratt et al. 2009).*
	
Usage:

.. function:: sd=scatpratt(x,y,errx,erry,a,b)

  :param x,y: X and Y data arrays
  :param errx, erry: standard deviations in X and Y
  :param a,b: slope and intercept of best-fit linear relation
  :rtype: float sd with the scatter about the best-fit.

v1 Mar 20 2012
	"""
	N=numpy.size(x)
	
	# Equation 4 from Pratt et al. 2009
	sdsq=erry**2+a**2*errx**2
	wden=1./N*numpy.sum(1./sdsq)	# Denominator of w
	w=1./sdsq/wden
	
	# Equation 3 from Pratt et al. 2009
	sdrawsq=1./(N-2.)*numpy.sum(w*(y-a*x-b)**2)
	sdraw=numpy.sqrt(sdrawsq)
	
	return sdraw



def residual(ydata,ymod):
	"""
Compute the residual sum of squares (RSS) also known as the sum of 
squared residuals (see http://en.wikipedia.org/wiki/Residual_sum_of_squares).

Usage:

>>> rss=residual(ydata,ymod)

where
  ydata : data
  ymod : model evaluated at xdata
  
ydata,ymod assumed to be Numpy arrays. 
Returns the float rss with the mean deviation.

Dec. 2011
	"""	
	# Std. deviation of an individual measurement (Bevington, eq. 6.15)
	N=numpy.size(ydata)
	rss=numpy.sum((ydata-ymod)**2)
	
	return rss



def chisq(x,y,a=None,b=None,sd=None):
	"""
Returns the chi-square error statistic as the sum of squared errors between
Y(i) and AX(i) + B. If individual standard deviations (array sd) are supplied, 
then the chi-square error statistic is computed as the sum of squared errors
divided by the standard deviations.	Inspired on the IDL procedure linfit.pro.
See http://en.wikipedia.org/wiki/Goodness_of_fit for reference.
	
If a linear model is not provided via A,B (y=ax+b) then the method computes
the chi-square using the best-fit line.

x,y,sd assumed to be Numpy arrays. a,b scalars.
Returns the float chisq with the chi-square statistic.
	"""

	if a==None:	
		# Performs linear regression
		a, b, r, p, err = scipy.stats.linregress(x,y)
	
	# Chi-square statistic (Bevington, eq. 6.9)
	if sd==None:
		chisq=numpy.sum((y-a*x-b)**2)
	else:
		chisq=numpy.sum( ((y-a*x-b)/sd)**2 )
	
	return chisq




def chisqg(ydata,ymod,sd=None):
	"""
Returns the chi-square error statistic as the sum of squared errors between
Ydata(i) and Ymodel(i). If individual standard deviations (array sd) are supplied, 
then the chi-square error statistic is computed as the sum of squared errors
divided by the standard deviations.	Inspired on the IDL procedure linfit.pro.
See http://en.wikipedia.org/wiki/Goodness_of_fit for reference.

x,y,sd assumed to be Numpy arrays. a,b scalars.
Returns the float chisq with the chi-square statistic.
	"""
	# Chi-square statistic (Bevington, eq. 6.9)
	if sd==None:
		chisq=numpy.sum((ydata-ymod)**2)
	else:
		chisq=numpy.sum( ((ydata-ymod)/sd)**2 )
	
	return chisq






def chisqxy(x,y,errx,erry,a,b):
	"""
Returns the chi-square error statistic for a linear fit, 
computed taking into account the errors in both X and Y (i.e. the
effective variance). See equation 3 in Ascenso et al. 2012, A&A or
Yee & Ellingson 2003 ApJ.

Usage:

>>> chisq=chisqxy(xdata,ydata,errx,erry,a,b)

where

- xdata,ydata : data
- errx,erry : measurement uncertainties in the data
- a,b : slope and intercept of the best-fit linear regression model
	"""
	sdsq=erry**2+a**2*errx**2
	chisq=numpy.sum( (y-a*x-b)**2/sdsq )
	
	return chisq





def intscat(x,y,errx,erry,a,b):
	"""
Estimates the intrinsic scatter about the best-fit, taking into account
the errors in X and Y and the "raw" scatter. Inspired by Pratt et al. 2009,
A&A, 498, 361. 
	
Usage:

>>> sd=intscat(x,y,errx,erry,a,b)

where
  
- x,y : X and Y data arrays
- errx, erry : standard deviations in X and Y
- a,b : slope and intercept of best-fit linear relation
  
Returns the float sd with the scatter about the best-fit.

v1 Apr 9th 2012
	"""
	# Raw scatter
	sdraw=scatpratt(x,y,errx,erry,a,b)
	sdrawsq=sdraw**2
	
	# Statistical variance, eq. 4 from Pratt et al. 2009
	sdsq=erry**2+a**2*errx**2
	
	# Intrinsic scatter for each data point
	intscati=numpy.sqrt( numpy.abs(sdsq-sdrawsq) )
	
	return numpy.mean(intscati)
	
	


def linregress_error(x,y):
	"""
Compute the uncertainties in the parameters A,B of the least-squares
linear regression fit y=ax+b to the data (scipy.stats.linregress).

x,y assumed to be Numpy arrays.
Returns the sequence sigma_a, sigma_b with the standard deviations in A and B.
	"""
	
	# Performs linear regression
	a, b, r, p, err = scipy.stats.linregress(x,y)
	
	# Std. deviation of an individual measurement (Bevington, eq. 6.15)
	N=numpy.size(x)
	sd=1./(N-2.)* numpy.sum((y-a*x-b)**2);	sd=numpy.sqrt(sd)
	
	# Std. deviation in parameters
	Delta=N*numpy.sum(x**2)-(numpy.sum(x))**2	# Bevington, eq. 6.13
	sdb=sd**2/Delta*numpy.sum(x**2);	sdb=numpy.sqrt(sdb)	# Bevington, eq. 6.23
	sda=N*sd**2/Delta;	sda=numpy.sqrt(sda)
	
	return sda,sdb
		
	










	


# Computing prediction and confidence bands
# ===========================================
#
#


def confband(xd,yd,a,b,conf=0.95,x=None):
	"""
Calculates the confidence band of the linear regression model at the desired confidence
level, using analytical methods. The 2sigma confidence interval is 95% sure to contain 
the best-fit regression line. This is not the same as saying it will contain 95% of 
the data points.

Arguments:

- conf: desired confidence level, by default 0.95 (2 sigma)
- xd,yd: data arrays
- a,b: linear fit parameters as in y=ax+b
- x: (optional) array with x values to calculate the confidence band. If none is provided, will by default generate 100 points in the original x-range of the data.
  
Returns:
Sequence (lcb,ucb,x) with the arrays holding the lower and upper confidence bands 
corresponding to the [input] x array.

Usage:

>>> lcb,ucb,x=nemmen.confband(all.kp,all.lg,a,b,conf=0.95)
calculates the confidence bands for the given input arrays

>>> pylab.fill_between(x, lcb, ucb, alpha=0.3, facecolor='gray')
plots a shaded area containing the confidence band

References:
1. http://en.wikipedia.org/wiki/Simple_linear_regression, see Section Confidence intervals
2. http://www.weibull.com/DOEWeb/confidence_intervals_in_simple_linear_regression.htm

v1 Dec. 2011
v2 Jun. 2012: corrected bug in computing dy
	"""
	alpha=1.-conf	# significance
	n=xd.size	# data sample size

	if x==None: x=numpy.linspace(xd.min(),xd.max(),100)

	# Predicted values (best-fit model)
	y=a*x+b

	# Auxiliary definitions
	sd=scatterfit(xd,yd,a,b)	# Scatter of data about the model
	sxd=numpy.sum((xd-xd.mean())**2)
	sx=(x-xd.mean())**2	# array

	# Quantile of Student's t distribution for p=1-alpha/2
	q=scipy.stats.t.ppf(1.-alpha/2.,n-2)

	# Confidence band
	dy=q*sd*numpy.sqrt( 1./n + sx/sxd )
	ucb=y+dy	# Upper confidence band
	lcb=y-dy	# Lower confidence band

	return lcb,ucb,x
	
	
	
	
	
def predband(xd,yd,a,b,conf=0.95,x=None):
	"""
Calculates the prediction band of the linear regression model at the desired confidence
level, using analytical methods. 

Clarification of the difference between confidence and prediction bands:
"The 2sigma confidence interval is 95% sure to contain the best-fit regression line. 
This is not the same as saying it will contain 95% of the data points. The prediction bands are
further from the best-fit line than the confidence bands, a lot further if you have many data 
points. The 95% prediction interval is the area in which you expect 95% of all data points to fall."
(from http://graphpad.com/curvefit/linear_regression.htm)

Arguments:

- conf: desired confidence level, by default 0.95 (2 sigma)
- xd,yd: data arrays
- a,b: linear fit parameters as in y=ax+b
- x: (optional) array with x values to calculate the confidence band. If none is provided, will
  by default generate 100 points in the original x-range of the data.
  
Usage:

>>> lpb,upb,x=nemmen.predband(all.kp,all.lg,a,b,conf=0.95)
calculates the prediction bands for the given input arrays

>>> pylab.fill_between(x, lpb, upb, alpha=0.3, facecolor='gray')
plots a shaded area containing the prediction band  

:returns: Sequence (lpb,upb,x) with the arrays holding the lower and upper confidence bands 
corresponding to the [input] x array.

References:

1. `Introduction to Simple Linear Regression, Gerard 
E. Dallal, Ph.D. <http://www.JerryDallal.com/LHSP/slr.htm>`_
	"""
	alpha=1.-conf	# significance
	n=xd.size	# data sample size

	if x is None: x=numpy.linspace(xd.min(),xd.max(),100)

	# Predicted values (best-fit model)
	y=a*x+b

	# Auxiliary definitions
	sd=scatterfit(xd,yd,a,b)	# Scatter of data about the model
	sxd=numpy.sum((xd-xd.mean())**2)
	sx=(x-xd.mean())**2	# array

	# Quantile of Student's t distribution for p=1-alpha/2
	q=scipy.stats.t.ppf(1.-alpha/2.,n-2)

	# Prediction band
	dy=q*sd*numpy.sqrt( 1.+1./n + sx/sxd )
	upb=y+dy	# Upper prediction band
	lpb=y-dy	# Lower prediction band

	return lpb,upb,x
	



def confbandnl(xd,yd,fun,par,varcov,deg,conf=0.95,x=None):
	"""
Calculates the confidence band of a nonlinear model at the desired confidence
level, using analytical methods. 

Arguments:

- xd,yd: data arrays
- fun : function f(v) - the model - which returns a scalar. v is an array
  such that v[0]=x (scalar), v[i>0] = parameters of the model
- par : array or list with structure [par0, par1, par2, ...] with the best-fit
  parameters that will be fed into fun
- varcov : variance-covariance matrix obtained from the nonlinear fit
- deg : number of free parameters in the model
- conf: desired confidence level, by default 0.95 (2 sigma)
- x: (optional) array with x values to calculate the confidence band. If none is provided, will
  by default generate 100 points in the original x-range of the data.
  
Usage:

>>> lcb,ucb,x=nemmen.confbandnl(all.kp,all.lg,broken,bfit,bcov,4,conf=0.95)
calculates the confidence bands for the given input arrays

>>> pylab.fill_between(x, lcb, ucb, alpha=0.3, facecolor='gray')
plots a shaded area containing the prediction band  

:returns: Sequence (lcb,ucb,x) with the arrays holding the lower and upper confidence bands 
corresponding to the [input] x array.

References:

1. `How does Prism compute confidence and prediction bands for nonlinear regression? <http://www.graphpad.com/faq/viewfaq.cfm?faq=1099>`_
2. http://stats.stackexchange.com/questions/15423/how-to-compute-prediction-bands-for-non-linear-regression
3. see also my notebook
	"""
	import numdifftools
	from . import misc
	
	alpha=1.-conf	# significance
	n=xd.size	# data sample size

	if x is None: x=numpy.linspace(xd.min(),xd.max(),100)

	# Gradient (needs to be evaluated)
	dfun=numdifftools.Gradient(fun)
	
	# Quantile of Student's t distribution for p=1-alpha/2
	q=scipy.stats.t.ppf(1.-alpha/2.,n-2)

	# Residual sum of squares		
	rss=residual(yd, misc.evalfun(fun,xd,par) )
	
	grad,p=[],[]
	i=0
	y=misc.evalfun(fun,x,par)
	v=numpy.zeros_like(x)
		
	for i in range(x.size):
		# List: arrays consisting of [x[i], par1, par2, ...]
		p.append( numpy.concatenate(([x[i]],par)) )
	
		# List: each element -> gradient evaluated at each xi
		grad.append(dfun(p[i]))
				
		# Before processing the grad array, eliminates the first element
		temp=numpy.dot(numpy.transpose(grad[i][1:]), varcov)
		v[i]=numpy.dot(temp, grad[i][1:])
		
	# Confidence band
	dy=q*numpy.sqrt( v*rss/(n-deg) )
	ucb=y+dy	# Upper confidence band
	lcb=y-dy	# Lower confidence band

	return lcb,ucb,x





def predbandnl(xd,yd,fun,par,varcov,deg,conf=0.95,x=None):
	"""
Calculates the prediction band of a nonlinear model at the desired confidence
level, using analytical methods. 

Arguments:

- xd,yd: data arrays
- fun : function f(v) - the model - which returns a scalar. v is an array
  such that v[0]=x, v[i>0] = parameters of the model
- par : array or list with structure [par0, par1, par2, ...] with the best-fit
  parameters that will be fed into fun
- varcov : variance-covariance matrix obtained from the nonlinear fit
- deg : number of free parameters in the model
- conf: desired confidence level, by default 0.95 (2 sigma)
- x: (optional) array with x values to calculate the confidence band. If none is provided, will
  by default generate 100 points in the original x-range of the data.
  
Usage:

>>> lpb,upb,x=nemmen.predbandnl(all.kp,all.lg,broken,bfit,bcov,4,conf=0.95)
calculates the prediction bands for the given input arrays

>>> pylab.fill_between(x, lpb, upb, alpha=0.3, facecolor='gray')
plots a shaded area containing the prediction band  

:returns: Sequence (lpb,upb,x) with the arrays holding the lower and upper confidence bands 
corresponding to the [input] x array.

References:
1. http://www.graphpad.com/faq/viewfaq.cfm?faq=1099, "How does Prism compute confidence and prediction bands for nonlinear regression?"
2. http://stats.stackexchange.com/questions/15423/how-to-compute-prediction-bands-for-non-linear-regression
3. see also my notebook)
	"""
	import numdifftools
	
	alpha=1.-conf	# significance
	n=xd.size	# data sample size

	if x is None: x=numpy.linspace(xd.min(),xd.max(),100)

	# Gradient (needs to be evaluated)
	dfun=numdifftools.Gradient(fun)
	
	# Quantile of Student's t distribution for p=1-alpha/2
	q=scipy.stats.t.ppf(1.-alpha/2.,n-2)

	# Residual sum of squares		
	rss=residual(yd, evalfun(fun,xd,par) )
	
	grad,p=[],[]
	i=0
	y=evalfun(fun,x,par)
	v=numpy.zeros_like(x)
		
	for i in range(x.size):
		# List: arrays consisting of [x[i], par1, par2, ...]
		p.append( numpy.concatenate(([x[i]],par)) )
	
		# List: each element -> gradient evaluated at each xi
		grad.append(dfun(p[i]))
				
		# Before processing the grad array, eliminates the first element
		temp=numpy.dot(numpy.transpose(grad[i][1:]), varcov)
		v[i]=numpy.dot(temp, grad[i][1:])
		
	# Confidence band
	dy=q*numpy.sqrt( (1.+v)*rss/(n-deg) )
	upb=y+dy	# Upper prediction band
	lpb=y-dy	# Lower prediction band

	return lpb,upb,x
		
	


def confbandmc(x,par,varcov,n=100000,sigmas=1.):
	"""
Calculates the 1sigma confidence band of a linear model without needing
to specify the data points, by doing a Monte Carlo simulation using the
Multivariate normal distribution. Assumes the parameters follow a 2D
normal distribution.

Arguments:

- x: array with x values to calculate the confidence band. 
- par : array or list with structure [par0, par1, par2, ...] with the best-fit
  parameters
- varcov : variance-covariance matrix of the parameters
- n : number of (a,b) points generated from the multivariate gaussian
- sigmas : number of standard deviations contained within the prediction
  band

Output:

- lcb, ucb : lower and upper prediction bands
- y : values of the model tabulated at x

Usage:

>>> lcb,ucb,y=nemmen.confbandmc(x,bfit,bcov)
calculates the prediction bands for the given input arrays

>>> pylab.fill_between(x, lcb, ucb, alpha=0.3, facecolor='gray')
plots a shaded area containing the prediction band
	"""
	# Generates many realizations of a and b from the multinormal distribution
	ar,br = numpy.random.multivariate_normal(par,varcov,n).T
	
	erry=[]	# will contain the std deviation in y
	y=[]	# values of y for each x
	
	for xi in x:
		yr=ar*xi+br
		erry.append( yr.std() )
		y.append( yr.mean() )	
	
	erry=numpy.array(erry)
	y=numpy.array(y)

	ucb=y+sigmas*erry	# Upper confidence band
	lcb=y-sigmas*erry	# Lower confidence band

	return lcb,ucb,y
	
	
	
	
def credbandmc(x,slope,inter,sigmas=1.):
	"""
Calculates the confidence (or credibility in case of Bayesian analysis) 
band of a linear regression model, given the posterior probability distributions
of the slope and intercept (presumably computed via Bayesian regression,
see bayeslin.pro). These distributions do not need to be normal.

Arguments:

- x: array with x values to calculate the confidence band. 
- slope, inter: posterior distributions of slope and intercept for linear
  regression
- sigmas : number of standard deviations contained within the confidence
  band

Output:

- lcb, ucb : lower and upper confidence bands
- y : best-fit values of the model tabulated at x

Usage:

>>> lcb,ucb,y=nemmen.predbandmc(x,a,b)
calculates the prediction bands for the given input arrays

>>> pylab.fill_between(x, lcb, ucb, alpha=0.3, facecolor='gray')
plots a shaded area containing the prediction band

v1 Jun. 2012: inspired by private communication with B. Kelly.
	"""
	a,b = slope,inter
	
	# Define the confidence/credibility interval
	conf=1.-scipy.special.erf(sigmas/numpy.sqrt(2.))
	
	lcb,ucb,ym=numpy.zeros_like(x),numpy.zeros_like(x),numpy.zeros_like(x)
	
	for i, xi in enumerate(x): 
		# Distribution of y for each x
		yp=a*xi+b	# 'p' as in posterior
		
		# Lower confidence band
		lcb[i]=scipy.stats.scoreatpercentile(yp,conf*100./2.)
		# Upper confidence band
		ucb[i]=scipy.stats.scoreatpercentile(yp,100.*(1-conf/2.))
		
		ym[i]=numpy.median(yp)	
		

	return lcb,ucb,ym
	
	
	

















# Monte Carlo simulations, generate random numbers
# ===================================================	
	

def gauss2d(par,varcov,n=10000):
	"""
Calculates random numbers drawn from the multinormal distribution in
two dimensions. Computes also the probability associated with each
number that can be used to computed confidence levels.

Arguments:

- par : array or list with structure [par0, par1, par2, ...] with the best-fit
  parameters
- varcov : variance-covariance matrix of the parameters
- n : number of (a,b) points generated from the multivariate gaussian

Output:

- x,y : arrays of random values generated from the multinormal distribution
- prob : probability associated with each value

Usage:
>>> x,y,prob=gauss2d(par,varcov,100000)

v1 Apr. 2012
	"""
	# Generates many realizations of a and b from the multinormal distribution
	x,y = numpy.random.multivariate_normal(par,varcov,n).T
	
	# Best-fit values
	x0,y0 = par[0], par[1]
	
	dx,dy = x-x0,y-y0
	sigx,sigy=numpy.sqrt(varcov[0,0]),numpy.sqrt(varcov[1,1])
	cov=varcov[0,1]
	rho=cov/sigx*sigy
	
	numchisq=(dx/sigx)**2+(dy/sigy)**2-2.*rho*(dx/sigx)*(dy/sigy)
	chisq=numchisq/(1.-rho**2)
	prob=numpy.exp(-chisq/2.)
	
	return x,y,prob




def bootcorr(x,y,nboot):
	"""
Given (X,Y) data points with intrinsic scatter, computes the 
bootstrapped Pearson and Spearman correlation coefficients. This
will give you an idea of how much outliers affect your correlation.

Usage:

>>> r,rho=bootcorr(x,y,100000)

performs 100000 bootstrapping realizations on the arrays x and y.

:returns: *r* - array with bootstrapped Pearson statistics
:returns: *rho* - bootstrapped array with Spearman statistics

	"""
	r,rho=[],[]
	for i in range(nboot):
		[xsim,ysim]=bootstrap([x,y])
		
		# Pearson r
		rsim=scipy.stats.pearsonr(xsim,ysim)	
		r.append(rsim[0])
		
		# Spearman rho
		rhosim=scipy.stats.spearmanr(xsim,ysim)
		rho.append(rhosim[0])

	r,rho=numpy.array(r),numpy.array(rho)

	results=numpy.array([ numpy.median(r), r.std(), numpy.median(rho), rho.std() ])
	print("<r>    err_r <rho> errrho")
	print(numpy.round(results, 2))

	results=numpy.array([ r2p(numpy.median(r)-numpy.abs(r.std()),x.size), r2p(numpy.median(r),x.size), r2p(numpy.median(r)+numpy.abs(r.std()),x.size) ])
	print("Prob. <- <r>-std,  <r>,    <r>+std")
	print(results)

	print("Rejection of H0 respectively at")
	for p in results:
		print(round(p2sig(p),2)	)
		
	return r,rho


def gen_ts(y,erry,n,zeropad=True):
    """
Given a time series (TS) with uncertainties on the signal, this will generate 
*n* new TS with y-values distributed according to the error bars. 

Output will be a :math:`n \\times {\rm{size}(t)}` array. Each row of this
array contains a simulated TS.

:param y: array of y-values for time series (do not need to be in order)
:param erry: array of 1-sigma errors on y-values
:param n: number of Mock TS to generate
:param zeropad: are y-values<0 not allowed? `True` will make any values<0 into 0
:returns: `n x size(t)` array. Each row of this array contains a simulated TS
    """
    ysim=numpy.empty((n,y.size))
    
    for i in range(y.size):
        # generate new points given normal distribution
        ysim[:,i]=numpy.random.normal(y[i],erry[i],n)   
        
    # makes sure no value is smaller than zero
    ysim[ysim<0]=0.
    
    return ysim


def random_normal(mean,std,n):
	"""
Returns an array of n elements of random variables, following a normal 
distribution with the supplied mean and standard deviation.

.. warning:: this is superseded. Use `numpy.random.normal <https://docs.scipy.org/doc/numpy/reference/generated/numpy.random.normal.html>`_ instead.
	"""
	return std*scipy.random.standard_normal(n)+mean


def random(a,b,n=None):
	"""
Generates an array of random uniformly distributed floats in the 
interval *[x0,x1)*.

>>> random(0.3,0.4,1000)
	"""
	return (b - a) * scipy.random.random_sample(n) + a

















	
	
	
# Comparing goodness-of-fit of different models
# ================================================
# F-test, Akaike information criterion and reduced chi-squared	

def ftest(rss1,rss2,n,p1,p2):
	"""
Carries out the F-test to decide which model fits the data better.
Computes the F statistic, associated p-value and the significance
in sigmas with which we can reject the null hypothesis. You can also
give the :math:`\chi^2` values instead of RSS if you have y-errors.

Note that p2>p1 must be obeyed, i.e. model 2 is "nested" within model 1.

Usage:

>>> fstat,pvalue,conf = ftest(rss1,rss2,n,p1,p2)

Arguments:
- rss1, rss2 : residual sum of squares for models 1 and 2
- n : sample size, i.e. number of data points
- p1, p2 : number of free parameters in the models

Returns:
- fstat : the F statistic
- pvalue : p-value associated with the F statistic
- conf : significance in sigmas with which we can reject the null hypothesis

References:
1. http://en.wikipedia.org/wiki/F-test
2. http://graphpad.com/curvefit/2_models__1_dataset.htm, Graphpad.com, Comparing the fits of two models (CurveFit.com)

v1 Dec. 2011
v2 Jan 16 2012: added comment regarding y-errors
"""
	fstat=scipy.stats.f_value(rss1,rss2,n-p1,n-p2)
	pvalue=1.-scipy.stats.f.cdf(fstat,p2-p1,n-p2)
	conf=numpy.sqrt(2.)*scipy.special.erfinv(1.-pvalue)
	
	return fstat, pvalue, conf
	
	



def aic(k,n,rss,errors=False):
	"""
Computes the Akaike information criterion which is "a measure of the relative goodness of
fit of a statistical model". Unlike the F-test, it does not assume that one model is 
a particular case of the more complex one. If errors==False then assume the
errors are the same for every data point. Otherwise, assumes you are providing
the chi-squared values instead of RSS.

Usage:

>>> aic = aic(k,n,rss)

:param rss: residual sum of squares for the model in case errors=False, otherwise assumes you are giving the chi-square values
:param n: sample size, i.e. number of data points
:param k: number of free parameters in the model
:returns: AIC statistic

References:

1. Documentation for Origin software on fit comparison: http://www.originlab.com/index.aspx?go=Products/Origin/DataAnalysis/CurveFitting/NonlinearFitting&pid=1195 (first heard about this test there)
2. http://en.wikipedia.org/wiki/Akaike_information_criterion

v1 Dec 2011
	"""
	if errors==False:
		# AIC assuming the errors are identical and given the residual sum of squares,
		# see http://en.wikipedia.org/wiki/Akaike_information_criterion#Relevance_to_chi-squared_fitting
		aicstat=n*numpy.log(rss/n)+2.*k
	else:
		# If you have different errors for the data points, it will assume rss=chisq
		aicstat=rss+2.*k
			
	# AICc = AIC with a correction for finite sample size
	aicc=aicstat+2.*k*(k+1.)/(n-k-1.)
	
	return aicc
	
	
	

def bic(k,n,rss,errors=False):
	"""
Computes the Bayesian information criterion which is "a criterion for model selection
among a finite set of models". From wikipedia: "The BIC [...] introduces a 
penalty term for the number of parameters in the model. The penalty term is larger 
in BIC than in AIC."

In order to use BIC to quantify the evidence against a specific model, check out
the section "use of BIC" in the presentation "171:290 Model Selection, Lecture VI: 
The Bayesian Information Criterion" by Joseph Cavanaugh 
(http://myweb.uiowa.edu/cavaaugh/ms_lec_6_ho.pdf).

If errors==False then assume the errors are the same for every data point. 
Otherwise, assumes you are providing the chi-squared values instead of RSS.

Usage:
>>> bic = bic(k,n,rss)

Arguments:
- rss : residual sum of squares for the model in case errors=False, otherwise assumes you are giving the chi-square values
- n : sample size, i.e. number of data points
- k : number of free parameters in the model

Returns: BIC statistic

References:
1. http://en.wikipedia.org/wiki/Bayesian_information_criterion
2. Model Selection, Lecture VI: The Bayesian Information Criterion" by Joseph Cavanaugh (http://myweb.uiowa.edu/cavaaugh/ms_lec_6_ho.pdf)

v1 Apr 2012
	"""
	if errors==False:
		# BIC assuming the errors are identical and given the residual sum of squares,
		# using the unbiased variance
		bicstat=n*numpy.log(rss/(n-1.))+k*numpy.log(n)
	else:
		# If you have different errors for the data points, it will assume rss=chisq
		bicstat=rss+k*numpy.log(n)
	
	return bicstat
	
	



def redchisq(x,y,a=None,b=None,sd=None):
	"""
Returns the reduced chi-square error statistic for a linear fit:
  chisq/nu
where nu is the number of degrees of freedom. If individual standard deviations 
(array sd) are supplied, then the chi-square error statistic is computed as the 
sum of squared errors divided by the standard deviations. 
See http://en.wikipedia.org/wiki/Goodness_of_fit for reference.
	
If a linear model is not provided via A,B (y=ax+b) then the method computes
the chi-square using the best-fit line.

x,y,sd assumed to be Numpy arrays. a,b scalars. deg integer.
Returns the float chisq/nu with the reduced chi-square statistic.
	"""

	if a is None:	
		# Performs linear regression
		a, b, r, p, err = scipy.stats.linregress(x,y)
	
	# Chi-square statistic
	if sd is None:
		chisq=numpy.sum((y-a*x-b)**2)
	else:
		chisq=numpy.sum( ((y-a*x-b)/sd)**2 )
		
	# Number of degrees of freedom assuming 2 free parameters
	nu=x.size-3
	
	return chisq/nu
	



def redchisqxy(x,y,errx,erry,a,b):
	"""
Returns the reduced chi-square error statistic for a linear fit :math:`\chi^2/\nu`
where nu is the number of degrees of freedom. The chi-square statistic is
computed taking into account the errors in both X and Y (i.e. the
effective variance). See equation 3 in Ascenso et al. 2012, A&A or
Yee & Ellingson 2003 ApJ.

Usage:

>>> chisq=redchisqxy(xdata,ydata,errx,erry,a,b)

where

- xdata,ydata : data
- errx,erry : measurement uncertainties in the data
- a,b : slope and intercept of the best-fit linear regression model
	"""
	sdsq=erry**2+a**2*errx**2
	chisq=numpy.sum( (y-a*x-b)**2/sdsq )
		
	# Number of degrees of freedom assuming 2 free parameters
	nu=x.size-3
	
	return chisq/nu





def redchisqg(ydata,ymod,deg=2,sd=None):
	"""
Returns the reduced chi-square error statistic for an arbitrary model, 
chisq/nu, where nu is the number of degrees of freedom. If individual 
standard deviations (array sd) are supplied, then the chi-square error 
statistic is computed as the sum of squared errors divided by the standard 
deviations. See http://en.wikipedia.org/wiki/Goodness_of_fit for reference.

ydata,ymod,sd assumed to be Numpy arrays. deg integer.

Usage:

>>> chisq=redchisqg(ydata,ymod,n,sd)

where

- ydata : data
- ymod : model evaluated at the same x points as ydata
- n : number of free parameters in the model
- sd : uncertainties in ydata
  	"""
	# Chi-square statistic
	if sd is None:
		chisq=numpy.sum((ydata-ymod)**2)
	else:
		chisq=numpy.sum( ((ydata-ymod)/sd)**2 )
		
	# Number of degrees of freedom assuming 2 free parameters
	nu=ydata.size-1-deg
	
	return chisq/nu	


def r2(ydata,ymod):
	"""
Computes the "coefficient of determination" R^2. According to wikipedia,
"It is the proportion of variability in a data set that is accounted for 
by the statistical model. It provides a measure of how well future outcomes 
are likely to be predicted by the model."

This method was inspired by http://stackoverflow.com/questions/3460357/calculating-the-coefficient-of-determination-in-python

References:
1. http://en.wikipedia.org/wiki/Coefficient_of_determination
2. http://stackoverflow.com/questions/3460357/calculating-the-coefficient-of-determination-in-python

v1 Apr 18th 2012
	"""
	ss_tot = numpy.sum( (ydata-ydata.mean())**2 )
	ss_err = numpy.sum( (ydata-ymod)**2 )
	
	return 1.-ss_err/ss_tot



def fitstats(xdata,ydata,errx,erry,a,b):
	"""
Computes different useful statistics for the given (X +- errX, Y +- errY)
arrays of data. Also quantifies the goodness-of-fit of the provided
linear fit with slope 'a' and intercept 'b'.

Usage:

>>> r,rho,rchisq,scat,iscat,r2=fitstats(xdata,ydata,errx,erry,a,b)

  :param xdata,ydata: data
  :param errx,erry: measurement uncertainties in the data
  :param a,b: slope and intercept of the best-fit linear regression model
  
Returns:

	- Pearson 'r'
	- Spearman 'rho'
	- Reduced chi-squared 'chi^2_nu'
	- raw scatter about the best-fit
	- intrinsic scatter
	- Coefficient of determination 'R^2'
	"""
	# Pearson r
	r=scipy.stats.pearsonr(xdata,ydata)	
	# Spearman rho
	rho=scipy.stats.spearmanr(xdata,ydata)	
	# Reduced chi-squared
	rchisq=redchisqxy(xdata,ydata,errx,erry,a,b)
	# raw scatter
	scat=scatpratt(xdata,ydata,errx,erry,a,b)
	# intrinsic scatter	
	iscat=intscat(xdata,ydata,errx,erry,a,b)	
	# Coefficient of determination R^2
	rtwo=r2(ydata,a*xdata+b)
	
	return r,rho,rchisq,scat,iscat,rtwo


















# Operations on statistical distributions
# =======================================
#

def splitstd(x):
	"""
Given the input distribution, this method computes two standard deviations:
the left and right side spreads. This is especially useful if you are dealing
with a split normal distribution from your data and you want to quantify the
lower and upper error bars without having to fit a split normal distribution.

This is a quick nonparametric method, as opposed to the more computationally
intensive minimization involved when fitting a function.

Algorithm:

- divide the original distribution in two distributions at the mode
- create two new distributions, which are the symmetrical versions of the left and right side of the original one
- compute the standard deviations for the new mirror distributions

:param x: array or list containing the distribution
:returns: left.stddev, right.stddev

.. note:: do not forget to inspect *x*.
	"""
	#med=median(x)
	med=mode(x)

	x1=x[ numpy.where(x<med) ] # left side
	x1mirror=x1+2.*numpy.abs(x1-med) # mirror side
	x1new=numpy.concatenate((x1,x1mirror))

	x2=x[ numpy.where(x>=med) ] # right side
	x2mirror=x2-2.*numpy.abs(x2-med) # mirror side
	x2new=numpy.concatenate((x2,x2mirror))

	return x1new.std(), x2new.std()




def mode(x,**kwargs):
	"""
Finds the mode of a distribution, i.e. the value where the PDF peaks.

:param x: input list/array with the distribution
	"""
	yh,xh=numpy.histogram(x,50,normed=True,**kwargs)
	dxh=(xh[1]-xh[0])/2.
	xh=xh+dxh

	return xh[ search(yh.max(),yh) ]















# Custom statistical distributions
# ==================================
#

def randomvariate(pdf,n=1000,xmin=0,xmax=1):
	"""
Rejection method for random number generation:
Uses the rejection method for generating random numbers derived from an arbitrary 
probability distribution. For reference, see Bevington's book, page 84. Based on
rejection*.py.

Usage:

>>> randomvariate(P,N,xmin,xmax)

  where

:param P: probability distribution function from which you want to generate random numbers
:param N: desired number of random values
:param xmin,xmax: range of random numbers desired
  
:returns: the sequence (ran,ntrials) where
  	ran : array of shape N with the random variates that follow the input P
  	ntrials : number of trials the code needed to achieve N

Here is the algorithm:

- generate x' in the desired range
- generate y' between Pmin and Pmax (Pmax is the maximal value of your pdf)
- if y'<P(x') accept x', otherwise reject
- repeat until desired number is achieved

v1 Nov. 2011
	"""
	# Calculates the minimal and maximum values of the PDF in the desired
	# interval. The rejection method needs these values in order to work
	# properly.
	x=numpy.linspace(xmin,xmax,1000)
	y=pdf(x)
	pmin=0.
	pmax=y.max()

	# Counters
	naccept=0
	ntrial=0

	# Keeps generating numbers until we achieve the desired n
	ran=[]	# output list of random numbers
	while naccept<n:
		x=numpy.random.uniform(xmin,xmax)	# x'
		y=numpy.random.uniform(pmin,pmax)	# y'

		if y<pdf(x):
			ran.append(x)
			naccept=naccept+1
		ntrial=ntrial+1
	
	ran=numpy.asarray(ran)
	
	return ran,ntrial





def splitnorm(x,sig1=1.,sig2=1.,mu=0.):
	"""
Split normal distribution (or asymmetric gaussian) PDF. Useful when dealing
with asymmetric error bars.

See http://en.wikipedia.org/wiki/Split_normal_distribution.

:param x: input array where the PDF will be computed
:param sig1: left standard deviation
:param sig2: right std. dev.
:param mu: mode
:returns: probability distribution function for the array x
	"""
	const=numpy.sqrt(2./numpy.pi)/(sig1 + sig2)
    
	p=numpy.where(x<mu, 
		numpy.exp(-((x-mu)**2/(2.*sig1**2)))*const, 
		numpy.exp(-((x-mu)**2/(2.*sig2**2)))*const)
    
	return p



class splitnorm_gen(scipy.stats.rv_continuous):
    """
Split normal distribution defined using scipy.stats. Can be called as 
any scipy.stats distribution. I took the effort of defining the extra
CDF and PPF methods below in order to speedup calls to this class 
(I have particularly in mind the package *mcerp*).

:param x: input array where the PDF will be computed
:param sig1: left standard deviation
:param sig2: right std. dev.
:param mu: mode

Examples: 

Defines distribution with sig1=1, sig2=3, mu=0:

>>> split = splitnorm_gen(name='splitnorm', shapes='sig1, sig2, mu')
>>> s=split(1,3,0.0001)
>>> x=numpy.linspace(-10,10,100)

Computes PDF:

>>> s.pdf(x)

Computes CDF:

>>> s.cdf(x)

Generates 100 random numbers:

>>> s.rvs(100)

.. warning:: for some reason, this fails if mu=0.
    """
    def _pdf(self, x, sig1, sig2, mu):
        const=numpy.sqrt(2./numpy.pi)/(sig1 + sig2)
            
        p=numpy.where(x<=mu, 
                numpy.exp(-((x-mu)**2/(2.*sig1**2)))*const, 
                numpy.exp(-((x-mu)**2/(2.*sig2**2)))*const)
                      
        return p
    
    def _cdf(self, x, sig1, sig2, mu):
        const=1./(sig1+sig2)
        
        c=numpy.where(x<=mu, 
                sig1*const*(1.+scipy.special.erf((x-mu)/(numpy.sqrt(2.)*sig1))), 
                const*(sig1+sig2*scipy.special.erf((x-mu)/(numpy.sqrt(2.)*sig2)))
                )
        
        return c
    
    def _ppf(self,q,sig1,sig2,mu):
        nu=sig1/(sig1 + sig2)
        
        pf=numpy.where(q<=nu, 
                numpy.sqrt(2.)*sig1*scipy.special.erfinv((sig1 + sig2)/sig1*q - 1.) + mu, 
                numpy.sqrt(2)*sig2*scipy.special.erfinv(((sig1 + sig2)*q - sig1)/sig2) + mu
                )
        
        return pf     
    
    def _stats(self, sig1, sig2, mu):
        mean=mu+numpy.sqrt(2./numpy.pi)*(sig2 - sig1)
        var=(1.-2./numpy.pi)*(sig2-sig1)**2 + sig1*sig2
        skew=numpy.sqrt(2./numpy.pi)*(sig2-sig1)*( (4./numpy.pi-1.)*(sig2-sig1)**2+sig1*sig2 )
        kurt=None
        
        return mean, var, skew, kurt






















# p-values and significance
# ===========================
#


def r2p(r,n):
	"""
Given a value of the Pearson r coefficient, this method gives the
corresponding p-value of the null hypothesis (i.e. no correlation).

Code copied from scipy.stats.pearsonr source.

Usage:

>>> p=r2p(r,n)

where 'r' is the Pearson coefficient and n is the number of data
points.

:returns: p-value of the null hypothesis.
	"""
	df = n-2
	if abs(r) == 1.0:
		prob = 0.0
	else:
		t_squared = r*r * (df / ((1.0 - r) * (1.0 + r)))
		prob = scipy.stats.betai(0.5*df, 0.5, df / (df + t_squared))
        
	return prob



def p2sig(p):
	"""
Given the p-value Pnull (i.e. probability of the null hypothesis) below, 
evaluates the confidence level with which we can reject the hypothesis in standard 
deviations.

Inspired on prob2sig.nb.

Usage:

>>> s=p2sig(0.001)
	"""
	sig=numpy.sqrt(2.)*scipy.special.erfinv(1.-p)
	
	return sig



def conf2sig(p):
	"""
Given a confidence p-value, translates it into standard deviations.
E.g., p=0.683 -> 1sigma, p=0.954 -> 2sigma, etc.

Inspired on p2sig method.

Usage:

>>> s=conf2sig(0.683)
	"""
	sig=numpy.sqrt(2.)*scipy.special.erfinv(p)
	
	return sig



def sig2conf(sig):
	"""
Given a number of std. deviations, translates it into a p-value.
E.g., 1sigma -> p=0.683, 2sigma -> p=0.954, etc.

Inspired on p2sig method.

Usage:

>>> s=sig2conf(5.)
	"""
	return scipy.special.erf(sig/numpy.sqrt(2.))
	

	


"""
Peak detection in time series
===================================

Originally downloaded from https://gist.github.com/sixtenbe. 
However, now I can't find the original source for the code anymore.
"""

import numpy as np
from math import pi, log
import pylab
from scipy import fft, ifft
from scipy.optimize import curve_fit

#i = 10000
#x = np.linspace(0, 3.5 * pi, i)
#y = (0.3*np.sin(x) + np.sin(1.3 * x) + 0.9 * np.sin(4.2 * x) + 0.06*np.random.randn(i))


def _datacheck_peakdetect(x_axis, y_axis):
    if x_axis is None:
        x_axis = range(len(y_axis))
    
    if len(y_axis) != len(x_axis):
        raise ValueError('Input vectors y_axis and x_axis must have same length')
    
    #needs to be a numpy array
    y_axis = np.array(y_axis)
    x_axis = np.array(x_axis)
    return x_axis, y_axis
    
def _peakdetect_parabole_fitter(raw_peaks, x_axis, y_axis, points):
    """
    Performs the actual parabole fitting for the peakdetect_parabole function.
    
    keyword arguments:
    raw_peaks -- A list of either the maximium or the minimum peaks, as given
        by the peakdetect_zero_crossing function, with index used as x-axis
    x_axis -- A numpy list of all the x values
    y_axis -- A numpy list of all the y values
    points -- How many points around the peak should be used during curve
        fitting, must be odd.
    
    return -- A list giving all the peaks and the fitted waveform, format:
        [[x, y, [fitted_x, fitted_y]]]
        
    """
    func = lambda x, k, tau, m: k * ((x - tau) ** 2) + m
    fitted_peaks = []
    for peak in raw_peaks:
        index = peak[0]
        x_data = x_axis[index - points // 2: index + points // 2 + 1]
        y_data = y_axis[index - points // 2: index + points // 2 + 1]
        # get a first approximation of tau (peak position in time)
        tau = x_axis[index]
        # get a first approximation of peak amplitude
        m = peak[1]
        
        # build list of approximations
        # k = -m as first approximation?
        p0 = (-m, tau, m)
        popt, pcov = curve_fit(func, x_data, y_data, p0)
        # retrieve tau and m i.e x and y value of peak
        x, y = popt[1:3]
        
        # create a high resolution data set for the fitted waveform
        x2 = np.linspace(x_data[0], x_data[-1], points * 10)
        y2 = func(x2, *popt)
        
        fitted_peaks.append([x, y, [x2, y2]])
        
    return fitted_peaks
    
    
def peakdetect(y_axis, x_axis = None, lookahead = 200, delta=0):
    """
    Converted from/based on a MATLAB script at: 
    http://billauer.co.il/peakdet.html
    
    function for detecting local maximas and minmias in a signal.
    Discovers peaks by searching for values which are surrounded by lower
    or larger values for maximas and minimas respectively
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find peaks
    x_axis -- (optional) A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the postion of the peaks. If
        omitted an index of the y_axis is used. (default: None)
    lookahead -- (optional) distance to look ahead from a peak candidate to
        determine if it is the actual peak (default: 200) 
        '(sample / period) / f' where '4 >= f >= 1.25' might be a good value
    delta -- (optional) this specifies a minimum difference between a peak and
        the following points, before a peak may be considered a peak. Useful
        to hinder the function from picking up false peaks towards to end of
        the signal. To work well delta should be set to delta >= RMSnoise * 5.
        (default: 0)
            delta function causes a 20% decrease in speed, when omitted
            Correctly used it can double the speed of the function
    
    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tupple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*tab)
    """
    max_peaks = []
    min_peaks = []
    dump = []   #Used to pop the first hit which almost always is false
       
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # store data length for later use
    length = len(y_axis)
    
    
    #perform some checks
    if lookahead < 1:
        raise ValueError("Lookahead must be '1' or above in value")
    if not (np.isscalar(delta) and delta >= 0):
        raise ValueError("delta must be a positive number")
    
    #maxima and minima candidates are temporarily stored in
    #mx and mn respectively
    mn, mx = np.Inf, -np.Inf
    
    #Only detect peak if there is 'lookahead' amount of points after it
    for index, (x, y) in enumerate(zip(x_axis[:-lookahead], 
                                        y_axis[:-lookahead])):
        if y > mx:
            mx = y
            mxpos = x
        if y < mn:
            mn = y
            mnpos = x
        
        ####look for max####
        if y < mx-delta and mx != np.Inf:
            #Maxima peak candidate found
            #look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index+lookahead].max() < mx:
                max_peaks.append([mxpos, mx])
                dump.append(True)
                #set algorithm to only find minima now
                mx = np.Inf
                mn = np.Inf
                if index+lookahead >= length:
                    #end is within lookahead no more peaks can be found
                    break
                continue
            #else:  #slows shit down this does
            #    mx = ahead
            #    mxpos = x_axis[np.where(y_axis[index:index+lookahead]==mx)]
        
        ####look for min####
        if y > mn+delta and mn != -np.Inf:
            #Minima peak candidate found 
            #look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index+lookahead].min() > mn:
                min_peaks.append([mnpos, mn])
                dump.append(False)
                #set algorithm to only find maxima now
                mn = -np.Inf
                mx = -np.Inf
                if index+lookahead >= length:
                    #end is within lookahead no more peaks can be found
                    break
            #else:  #slows shit down this does
            #    mn = ahead
            #    mnpos = x_axis[np.where(y_axis[index:index+lookahead]==mn)]
    
    
    #Remove the false hit on the first value of the y_axis
    try:
        if dump[0]:
            max_peaks.pop(0)
        else:
            min_peaks.pop(0)
        del dump
    except IndexError:
        #no peaks were found, should the function return empty lists?
        pass
        
    return [max_peaks, min_peaks]
    
    
def peakdetect_fft(y_axis, x_axis, pad_len = 5):
    """
    Performs a FFT calculation on the data and zero-pads the results to
    increase the time domain resolution after performing the inverse fft and
    send the data to the 'peakdetect' function for peak 
    detection.
    
    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as the index 50.234 or similar.
    
    Will find at least 1 less peak then the 'peakdetect_zero_crossing'
    function, but should result in a more precise value of the peak as
    resolution has been increased. Some peaks are lost in an attempt to
    minimize spectral leakage by calculating the fft between two zero
    crossings for n amount of signal periods.
    
    The biggest time eater in this function is the ifft and thereafter it's
    the 'peakdetect' function which takes only half the time of the ifft.
    Speed improvementd could include to check if 2**n points could be used for
    fft and ifft or change the 'peakdetect' to the 'peakdetect_zero_crossing',
    which is maybe 10 times faster than 'peakdetct'. The pro of 'peakdetect'
    is that it resutls in one less lost peak. It should also be noted that the
    time used by the ifft function can change greatly depending on the input.
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the postion of the peaks.
    pad_len -- (optional) By how many times the time resolution should be
        increased by, e.g. 1 doubles the resolution. The amount is rounded up
        to the nearest 2 ** n amount (default: 5)
    
    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tupple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*tab)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    zero_indices = zero_crossings(y_axis, window = 11)
    #select a n amount of periods
    last_indice = - 1 - (1 - len(zero_indices) & 1)
    # Calculate the fft between the first and last zero crossing
    # this method could be ignored if the begining and the end of the signal
    # are discardable as any errors induced from not using whole periods
    # should mainly manifest in the beginning and the end of the signal, but
    # not in the rest of the signal
    fft_data = fft(y_axis[zero_indices[0]:zero_indices[last_indice]])
    padd = lambda x, c: x[:len(x) // 2] + [0] * c + x[len(x) // 2:]
    n = lambda x: int(log(x)/log(2)) + 1
    # padds to 2**n amount of samples
    fft_padded = padd(list(fft_data), 2 ** 
                n(len(fft_data) * pad_len) - len(fft_data))
    
    # There is amplitude decrease directly proportional to the sample increase
    sf = len(fft_padded) / float(len(fft_data))
    # There might be a leakage giving the result an imaginary component
    # Return only the real component
    y_axis_ifft = ifft(fft_padded).real * sf #(pad_len + 1)
    x_axis_ifft = np.linspace(
                x_axis[zero_indices[0]], x_axis[zero_indices[last_indice]],
                len(y_axis_ifft))
    # get the peaks to the interpolated waveform
    max_peaks, min_peaks = peakdetect(y_axis_ifft, x_axis_ifft, 500,
                                    delta = abs(np.diff(y_axis).max() * 2))
    #max_peaks, min_peaks = peakdetect_zero_crossing(y_axis_ifft, x_axis_ifft)
    
    # store one 20th of a period as waveform data
    data_len = int(np.diff(zero_indices).mean()) / 10
    data_len += 1 - data_len & 1
    
    
    fitted_wave = []
    for peaks in [max_peaks, min_peaks]:
        peak_fit_tmp = []
        index = 0
        for peak in peaks:
            index = np.where(x_axis_ifft[index:]==peak[0])[0][0] + index
            x_fit_lim = x_axis_ifft[index - data_len // 2:
                                    index + data_len // 2 + 1]
            y_fit_lim = y_axis_ifft[index - data_len // 2:
                                    index + data_len // 2 + 1]
            
            peak_fit_tmp.append([x_fit_lim, y_fit_lim])
        fitted_wave.append(peak_fit_tmp)
    
    #pylab.plot(range(len(fft_data)), fft_data)
    #pylab.show()
    
    pylab.plot(x_axis, y_axis)
    pylab.hold(True)
    pylab.plot(x_axis_ifft, y_axis_ifft)
    #for max_p in max_peaks:
    #    pylab.plot(max_p[0], max_p[1], 'xr')
    pylab.show()
    return [max_peaks, min_peaks]
    
    
def peakdetect_parabole(y_axis, x_axis, points = 9):
    """
    Function for detecting local maximas and minmias in a signal.
    Discovers peaks by fitting the model function: y = k (x - tau) ** 2 + m
    to the peaks. The amount of points used in the fitting is set by the
    points argument.
    
    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as index 50.234 or similar.
    
    will find the same amount of peaks as the 'peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the postion of the peaks.
    points -- (optional) How many points around the peak should be used during
        curve fitting, must be odd (default: 9)
    
    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a list
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # make the points argument odd
    points += 1 - points % 2
    #points += 1 - int(points) & 1 slower when int conversion needed
    
    # get raw peaks
    max_raw, min_raw = peakdetect_zero_crossing(y_axis)
    
    # define output variable
    max_peaks = []
    min_peaks = []
    
    max_ = _peakdetect_parabole_fitter(max_raw, x_axis, y_axis, points)
    min_ = _peakdetect_parabole_fitter(min_raw, x_axis, y_axis, points)
    
    max_peaks = map(lambda x: [x[0], x[1]], max_)
    max_fitted = map(lambda x: x[-1], max_)
    min_peaks = map(lambda x: [x[0], x[1]], min_)
    min_fitted = map(lambda x: x[-1], min_)
    
    
    #pylab.plot(x_axis, y_axis)
    #pylab.hold(True)
    #for max_p, max_f in zip(max_peaks, max_fitted):
    #    pylab.plot(max_p[0], max_p[1], 'x')
    #    pylab.plot(max_f[0], max_f[1], 'o', markersize = 2)
    #for min_p, min_f in zip(min_peaks, min_fitted):
    #    pylab.plot(min_p[0], min_p[1], 'x')
    #    pylab.plot(min_f[0], min_f[1], 'o', markersize = 2)
    #pylab.show()
    
    return [max_peaks, min_peaks]
    

def peakdetect_sine(y_axis, x_axis, points = 9, lock_frequency = False):
    """
    Function for detecting local maximas and minmias in a signal.
    Discovers peaks by fitting the model function:
    y = A * sin(2 * pi * f * x - tau) to the peaks. The amount of points used
    in the fitting is set by the points argument.
    
    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as index 50.234 or similar.
    
    will find the same amount of peaks as the 'peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.
    
    The function might have some problems if the sine wave has a
    non-negligible total angle i.e. a k*x component, as this messes with the
    internal offset calculation of the peaks, might be fixed by fitting a 
    k * x + m function to the peaks for offset calculation.
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the postion of the peaks.
    points -- (optional) How many points around the peak should be used during
        curve fitting, must be odd (default: 9)
    lock_frequency -- (optional) Specifies if the frequency argument of the
        model function should be locked to the value calculated from the raw
        peaks or if optimization process may tinker with it. (default: False)
    
    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tupple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*tab)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # make the points argument odd
    points += 1 - points % 2
    #points += 1 - int(points) & 1 slower when int conversion needed
    
    # get raw peaks
    max_raw, min_raw = peakdetect_zero_crossing(y_axis)
    
    # define output variable
    max_peaks = []
    min_peaks = []
    
    # get global offset
    offset = np.mean([np.mean(max_raw, 0)[1], np.mean(min_raw, 0)[1]])
    # fitting a k * x + m function to the peaks might be better
    #offset_func = lambda x, k, m: k * x + m
    
    # calculate an approximate frequenzy of the signal
    Hz = []
    for raw in [max_raw, min_raw]:
        if len(raw) > 1:
            peak_pos = [x_axis[index] for index in zip(*raw)[0]]
            Hz.append(np.mean(np.diff(peak_pos)))
    Hz = 1 / np.mean(Hz)
    
    # model function
    # if cosine is used then tau could equal the x position of the peak
    # if sine were to be used then tau would be the first zero crossing
    if lock_frequency:
        func = lambda x, A, tau: A * np.sin(2 * pi * Hz * (x - tau) + pi / 2)
    else:
        func = lambda x, A, Hz, tau: A * np.sin(2 * pi * Hz * (x - tau) + 
                                                pi / 2)
    #func = lambda x, A, Hz, tau: A * np.cos(2 * pi * Hz * (x - tau))
    
    
    #get peaks
    fitted_peaks = []
    for raw_peaks in [max_raw, min_raw]:
        peak_data = []
        for peak in raw_peaks:
            index = peak[0]
            x_data = x_axis[index - points // 2: index + points // 2 + 1]
            y_data = y_axis[index - points // 2: index + points // 2 + 1]
            # get a first approximation of tau (peak position in time)
            tau = x_axis[index]
            # get a first approximation of peak amplitude
            A = peak[1]
            
            # build list of approximations
            if lock_frequency:
                p0 = (A, tau)
            else:
                p0 = (A, Hz, tau)
            
            # subtract offset from waveshape
            y_data -= offset
            popt, pcov = curve_fit(func, x_data, y_data, p0)
            # retrieve tau and A i.e x and y value of peak
            x = popt[-1]
            y = popt[0]
            
            # create a high resolution data set for the fitted waveform
            x2 = np.linspace(x_data[0], x_data[-1], points * 10)
            y2 = func(x2, *popt)
            
            # add the offset to the results
            y += offset
            y2 += offset
            y_data += offset
            
            peak_data.append([x, y, [x2, y2]])
       
        fitted_peaks.append(peak_data)
    
    # structure date for output
    max_peaks = map(lambda x: [x[0], x[1]], fitted_peaks[0])
    max_fitted = map(lambda x: x[-1], fitted_peaks[0])
    min_peaks = map(lambda x: [x[0], x[1]], fitted_peaks[1])
    min_fitted = map(lambda x: x[-1], fitted_peaks[1])
    
    
    #pylab.plot(x_axis, y_axis)
    #pylab.hold(True)
    #for max_p, max_f in zip(max_peaks, max_fitted):
    #    pylab.plot(max_p[0], max_p[1], 'x')
    #    pylab.plot(max_f[0], max_f[1], 'o', markersize = 2)
    #for min_p, min_f in zip(min_peaks, min_fitted):
    #    pylab.plot(min_p[0], min_p[1], 'x')
    #    pylab.plot(min_f[0], min_f[1], 'o', markersize = 2)
    #pylab.show()
    
    return [max_peaks, min_peaks]

    
def peakdetect_sine_locked(y_axis, x_axis, points = 9):
    """
    Convinience function for calling the 'peakdetect_sine' function with
    the lock_frequency argument as True.
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the postion of the peaks.
    points -- (optional) How many points around the peak should be used during
        curve fitting, must be odd (default: 9)
        
    return -- see 'peakdetect_sine'
    """
    return peakdetect_sine(y_axis, x_axis, points, True)
    
    
def peakdetect_zero_crossing(y_axis, x_axis = None, window = 11):
    """
    Function for detecting local maximas and minmias in a signal.
    Discovers peaks by dividing the signal into bins and retrieving the
    maximum and minimum value of each the even and odd bins respectively.
    Division into bins is performed by smoothing the curve and finding the
    zero crossings.
    
    Suitable for repeatable signals, where some noise is tolerated. Excecutes
    faster than 'peakdetect', although this function will break if the offset
    of the signal is too large. It should also be noted that the first and
    last peak will probably not be found, as this function only can find peaks
    between the first and last zero crossing.
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find peaks
    x_axis -- (optional) A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the postion of the peaks. If
        omitted an index of the y_axis is used. (default: None)
    window -- the dimension of the smoothing window; should be an odd integer
        (default: 11)
    
    return -- two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tupple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*tab)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    
    zero_indices = zero_crossings(y_axis, window = window)
    period_lengths = np.diff(zero_indices)
            
    bins_y = [y_axis[index:index + diff] for index, diff in 
        zip(zero_indices, period_lengths)]
    bins_x = [x_axis[index:index + diff] for index, diff in 
        zip(zero_indices, period_lengths)]
        
    even_bins_y = bins_y[::2]
    odd_bins_y = bins_y[1::2]
    even_bins_x = bins_x[::2]
    odd_bins_x = bins_x[1::2]
    hi_peaks_x = []
    lo_peaks_x = []
    
    #check if even bin contains maxima
    if abs(even_bins_y[0].max()) > abs(even_bins_y[0].min()):
        hi_peaks = [bin.max() for bin in even_bins_y]
        lo_peaks = [bin.min() for bin in odd_bins_y]
        # get x values for peak
        for bin_x, bin_y, peak in zip(even_bins_x, even_bins_y, hi_peaks):
            hi_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
        for bin_x, bin_y, peak in zip(odd_bins_x, odd_bins_y, lo_peaks):
            lo_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
    else:
        hi_peaks = [bin.max() for bin in odd_bins_y]
        lo_peaks = [bin.min() for bin in even_bins_y]
        # get x values for peak
        for bin_x, bin_y, peak in zip(odd_bins_x, odd_bins_y, hi_peaks):
            hi_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
        for bin_x, bin_y, peak in zip(even_bins_x, even_bins_y, lo_peaks):
            lo_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
    
    max_peaks = [[x, y] for x,y in zip(hi_peaks_x, hi_peaks)]
    min_peaks = [[x, y] for x,y in zip(lo_peaks_x, lo_peaks)]
    
    return [max_peaks, min_peaks]
        
    
def _smooth(x, window_len=11, window='hanning'):
    """
    smooth the data using a window of the requested size.
    
    This method is based on the convolution of a scaled window on the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window; should be an odd 
            integer
        window: the type of window from 'flat', 'hanning', 'hamming', 
            'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    t = linspace(-2,2,0.1)
    x = sin(t)+randn(len(t))*0.1
    y = _smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, 
    numpy.convolve, scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if a list instead of
    a string   
    """
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")
    
    if window_len<3:
        return x
    
    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is not one of '{0}', '{1}', '{2}', '{3}', '{4}'".format(
            *('flat', 'hanning', 'hamming', 'bartlett', 'blackman')))
    
    s = np.r_[x[window_len-1:0:-1], x, x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w = np.ones(window_len,'d')
    else:
        w = eval('np.' + window + '(window_len)')

    y = np.convolve(w / w.sum(), s, mode = 'valid')
    return y
    
    
def zero_crossings(y_axis, window = 11):
    """
    Algorithm to find zero crossings. Smoothens the curve and finds the
    zero-crossings by looking for a sign change.
    
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find zero-crossings
    window -- the dimension of the smoothing window; should be an odd integer
        (default: 11)
    
    return -- the index for each zero-crossing
    """
    # smooth the curve
    length = len(y_axis)
    x_axis = np.asarray(range(length), int)
    
    # discard tail of smoothed signal
    y_axis = _smooth(y_axis, window)[:length]
    zero_crossings = np.where(np.diff(np.sign(y_axis)))[0]
    indices = [x_axis[index] for index in zero_crossings]
    
    # check if zero-crossings are valid
    diff = np.diff(indices)
    if diff.std() / diff.mean() > 0.2:
        print(diff.std() / diff.mean())
        print(np.diff(indices))
        raise ValueError("False zero-crossings found, indicates problem {0} or {1}".format(
            "with smoothing window", "problem with offset"))
    # check if any zero crossings were found
    if len(zero_crossings) < 1:
        raise ValueError("No zero crossings found")
    
    return indices 
    # used this to test the fft function's sensitivity to spectral leakage
    #return indices + np.asarray(30 * np.random.randn(len(indices)), int)
    
############################Frequency calculation#############################
#    diff = np.diff(indices)
#    time_p_period = diff.mean()
#    
#    if diff.std() / time_p_period > 0.1:
#        raise ValueError, 
#            "smoothing window too small, false zero-crossing found"
#    
#    #return frequency
#    return 1.0 / time_p_period
##############################################################################



    
    
def _test_zero():
    _max, _min = peakdetect_zero_crossing(y,x)
def _test():
    _max, _min = peakdetect(y,x, delta=0.30)
    
    
def _test_graph():
    i = 10000
    x = np.linspace(0,3.7*pi,i)
    y = (0.3*np.sin(x) + np.sin(1.3 * x) + 0.9 * np.sin(4.2 * x) + 0.06 *
    np.random.randn(i))
    y *= -1
    x = range(i)
    
    _max, _min = peakdetect(y,x,750, 0.30)
    xm = [p[0] for p in _max]
    ym = [p[1] for p in _max]
    xn = [p[0] for p in _min]
    yn = [p[1] for p in _min]
    
    plot = pylab.plot(x,y)
    pylab.hold(True)
    pylab.plot(xm, ym, 'r+')
    pylab.plot(xn, yn, 'g+')
    
    _max, _min = peak_det_bad.peakdetect(y, 0.7, x)
    xm = [p[0] for p in _max]
    ym = [p[1] for p in _max]
    xn = [p[0] for p in _min]
    yn = [p[1] for p in _min]
    pylab.plot(xm, ym, 'y*')
    pylab.plot(xn, yn, 'k*')
    pylab.show()
    
    
    
if __name__ == "__main__":
    from math import pi
    import pylab
    
    i = 10000
    x = np.linspace(0,3.7*pi,i)
    y = (0.3*np.sin(x) + np.sin(1.3 * x) + 0.9 * np.sin(4.2 * x) + 0.06 * 
    np.random.randn(i))
    y *= -1
    
    _max, _min = peakdetect(y, x, 750, 0.30)
    xm = [p[0] for p in _max]
    ym = [p[1] for p in _max]
    xn = [p[0] for p in _min]
    yn = [p[1] for p in _min]
    
    plot = pylab.plot(x, y)
    pylab.hold(True)
    pylab.plot(xm, ym, 'r+')
    pylab.plot(xn, yn, 'g+')
    
    
    pylab.show()


from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='nmmn',
    version='0.8.1',
    description='Miscellaneous methods for astronomy, dealing with arrays, statistical distributions and computing goodness-of-fit',
    long_description=readme,
    author='Rodrigo Nemmen',
    author_email='rodrigo.nemmen@iag.usp.br',
    url='https://github.com/rsnemmen/nmmn',
    download_url = 'https://github.com/peterldowns/mypackage/archive/0.8.1.tar.gz', # I'll explain this in a second
    license=license,
    keywords = ['science', 'statistics', 'signal-processing', 'numerical-methods', 'astronomy', 'numerical-simulations', 'astrophysics', 'mhd', 'grmhd'], # arbitrary keywords
    packages=find_packages(exclude=('tests', 'docs'))
)




"""
Miscellaneous modules for:

* `astro`: astronomy
* `dsp`: signal processing
* `lsd`: misc. operations on arrays, lists, dictionaries and sets
* `stats`: statistical methods
* `plots`: custom plots
* `fermi`: Fermi LAT analysis methods
* `bayes`: Bayesian tools for dealing with posterior distributions
* `grmhd`: tools for dealing with GRMHD numerical simulations

These are modules I wrote which I find useful -- for whatever reason -- in my research.

.. moduleauthor:: Rodrigo Nemmen <rodrigo.nemmen@iag.usp.br>
"""


"""
Astrophysical routines
=========================

"""


import numpy
import scipy




def dist2z(d):
	"""
Converts luminosity distance to redshift by solving the equation
'd-z=0'.

Input is assumed float.
	"""
	import cosmolopy 

	# x here is the unknown redshift
	f = lambda x: d-cosmolopy.distance.luminosity_distance(x,**cosmolopy.fidcosmo)
 
	z = scipy.optimize.fsolve(f, 0.01)  
	return z

 


def mjy(lognu,ll,dist,llerr=None):
    """
Converts log(nu/Hz), log(nu Lnu [erg/s]), error in log(nuLnu) to
log(lambda/micron), log(Fnu/mJy), error in log(Fnu).
The input units are CGS.

Usage:
If you have errors in the flux:

>>> lamb,fnu,ferr=mjy(xdata,ydata,dist,yerr)

If you do not have errors in the flux:

>>> lamb,fnu=mjy(xdata,ydata,dist)

:param dist: distance in Mpc
    """
    import uncertainties.unumpy as unumpy

    c=29979245800.	# speed of light in CGS
    dist=dist*3.085677581e24	# Mpc -> cm

    nu=10**lognu
    lamb=c/nu*1e4 # cm -> micron
    if llerr!=None:
    	lllerr=unumpy.uarray(ll,llerr)
    else:
    	lllerr=ll
    lnuerr=10**lllerr/nu  
    fluxerr=lnuerr/(1e-26*4.*numpy.pi*dist**2) # Lnu (erg/s/Hz) -> Fnu (mJy)
    if llerr!=None:
    	fluxerr=unumpy.log10(fluxerr)
    	return numpy.log10(lamb),unumpy.nominal_values(fluxerr),unumpy.std_devs(fluxerr)
    else:
    	return numpy.log10(lamb),numpy.log10(fluxerr)






def arcsec2pc(d=15.,a=1.):
	"""
Given the input angular size and distance to the object, computes
the corresponding linear size in pc. 

:param d: distance in Mpc
:param a: angular size in arcsec
:returns: linear size in pc
	"""

	# convert arcsec to radians
	a=a*4.848e-6	
	# convert distance to pc instead of Mpc
	d=d*1e6

	return d*numpy.tan(a)



def freq(T):
    """
Convert array of periods in days to frequencies in Hz.
    """
    return 1./T/86400.



def period(freq):
    """
Convert array of frequencies to periods.
    """
    return 1./freq




class Constants:
    """
Defines a set of useful constants in CGS.
::

    const=nmmn.astro.Constants()
    E=mass*const.c**2
    """

    def __init__(self):
        self.G = 6.673E-8
        self.c = 29979245800
        self.solarmass = 1.99e33
        self.year = 31556926
        
